#
# Copyright 2008-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Recipe for the reduction of imaging mode observations."""

import datetime
import logging
import os
import shutil
import uuid

import numpy
import sep
from astropy.io import fits
from scipy import interpolate
from scipy.spatial import KDTree as KDTree

from numina.core import Result, Requirement, Parameter
from numina.core.requirements import ObservationResultRequirement
from numina.core.query import ResultOf
import numina.array as narray
import numina.array.utils as nautils
import numina.array.combine as nacom
import numina.frame.combine as nfcom
from numina.util.context import manage_fits
from numina.frame import resize_fits, custom_region_to_str

import emirdrp.requirements as reqs
import emirdrp.products as prods
from emirdrp.processing.wcs import offsets_from_wcs_imgs
from emirdrp.processing.corr import offsets_from_crosscor_regions
from emirdrp.core.recipe import EmirRecipe

from .naming import (name_redimensioned_frames, name_object_mask,
                     name_skybackground)
from .naming import name_skybackgroundmask, name_skysub_proc, name_skyflat
from .naming import name_skyflat_proc, name_segmask

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2


class ImageInfo(object):
    def __init__(self, origin):
        self.origin = origin
        self.mask = None
        self.metadata = {}
        self.objmask_data = None
        self.valid_target = False
        self.valid_sky = False
        self.label = ""
        self.itype = 'SKY'
        self.valid_region = Ellipsis
        self.resized_base = ""
        self.lastname = ""
        self.flat_corrected = ""

    def __str__(self):
        output = ''
        for item in vars(self):
            value = getattr(self, item)
            output += '\n- {}: {}'.format(item, value)
            if hasattr(value, 'shape'):
                output += ' --> shape {}'.format(value.shape)
        return output


class FullDitheredImagesRecipe(EmirRecipe):
    """Recipe for the reduction of imaging mode observations.

    Recipe to reduce observations obtained in imaging mode, considering
    different possibilities depending on the size of the offsets
    between individual images.
    In particular, the following observing modes are considered: stare imaging,
    nodded beamswitched imaging, and dithered imaging.

    A critical piece of information here is a table that clearly specifies
    which images can be labeled as *science*, and which ones as *sky*.
    Note that some images are used both as *science* and *sky*
    (when the size of the targets is small compared to the offsets).

    **Observing modes:**

     * StareImage
     * Nodded/Beam-switched images
     * Dithered images


    **Inputs:**

     * Science frames + [Sky Frames]
     * Observing mode name: **stare image**, **nodded beamswitched image**,
       or **dithered imaging**
     * A table relating each science image with its sky image(s) (TBD if
       it's in the FITS header and/or in other format)
     * Offsets between them (Offsets must be integer)
     * Master Dark
     * Bad pixel mask (BPM)
     * Non-linearity correction polynomials
     * Master flat (twilight/dome flats)
     * Master background (thermal background, only in K band)
     * Exposure Time (must be the same in all the frames)
     * Airmass for each frame
     * Detector model (gain, RN, lecture mode)
     * Average extinction in the filter
     * Astrometric calibration (TBD)

    **Outputs:**

     * Image with three extensions: final image scaled to the individual
       exposure time, variance  and exposure time map OR number of images
       combined (TBD)

    **Procedure:**

    Images are corrected from dark, non-linearity and flat. Then, an iterative
    process starts:

     * Sky is computed from each frame, using the list of sky images of each
       science frame. The objects are avoided using a mask (from the second
       iteration on).

     * The relative offsets are the nominal from the telescope. From the second
       iteration on, we refine them using objects of appropriate brightness
       (not too bright, not to faint).

     * We combine the sky-subtracted images, output is: a new image, a variance
       image and a exposure map/number of images used map.

     * An object mask is generated.

     * We recompute the sky map, using the object mask as an additional input.
       From here we iterate (typically 4 times).

     * Finally, the images are corrected from atmospheric extinction and flux
       calibrated.

     * A preliminary astrometric calibration can always be used (using
       the central coordinates of the pointing and the plate scale
       in the detector).
       A better calibration might be computed using available stars (TBD).

    """

    logger = logging.getLogger(__name__)

    obresult = ObservationResultRequirement(
        query_opts=ResultOf('result_image', node='children')
    )

    master_bpm = reqs.MasterBadPixelMaskRequirement()

    offsets = Requirement(
        prods.CoordinateList2DType,
        'List of pairs of offsets',
        optional=True
    )
    refine_offsets = Parameter(False, 'Refine offsets by cross-correlation')
    iterations = Parameter(0, 'Iterations of the recipe')
    extinction = Parameter(0.0, 'Mean atmospheric extinction')

    method = Parameter(
        'sigmaclip',
        description='Combination method',
        choices=['mean', 'median', 'sigmaclip']
    )
    method_kwargs = Parameter(
        dict(),
        description='Arguments for combination method',
        optional=True
    )

    sky_images = Parameter(
        0, 'Images used to estimate the '
           'background before and after current image')

    sky_images_sep_time = Parameter(
        10, 'Maximum time interval between target and sky images [minutes]'
    )

    nside_adhoc_sky_correction = Parameter(
        0, 'Ad hoc sky correction (number of subintervals in each quadrant)'
    )

    result_image = Result(prods.ProcessedImage)
    result_sky = Result(prods.ProcessedImage, optional=True)

    def run(self, rinput):

        target_is_sky = True
        obresult = rinput.obresult
        sky_images = rinput.sky_images
        sky_images_sep_time = rinput.sky_images_sep_time
        baseshape = (EMIR_NAXIS2, EMIR_NAXIS1)
        user_offsets = rinput.offsets
        extinction = rinput.extinction
        nside_adhoc_sky_correction = rinput.nside_adhoc_sky_correction

        # protections
        if rinput.iterations == 0 and sky_images != 0:
            raise ValueError('sky_images: {} not compatible with iterations: {}'.format(
                sky_images, rinput.iterations
            ))

        if rinput.iterations > 0 and sky_images == 0:
            raise ValueError('iterations != 0 requires sky_images > 0')

        # check combination method
        if rinput.method != 'sigmaclip':
            if rinput.method_kwargs != {}:
                raise ValueError('Unexpected method_kwargs={}'.format(
                    rinput.method_kwargs))
        # combination method and arguments
        method = getattr(nacom, rinput.method)
        method_kwargs = rinput.method_kwargs

        images_info = self.initial_classification(obresult, target_is_sky)

        # Resizing target frames
        target_info = [iinfo for iinfo in images_info if iinfo.valid_target]
        finalshape, offsetsp, refpix, offset_fc0 = self.compute_size(
            target_info, baseshape, user_offsets
        )

        self.resize(target_info, baseshape, offsetsp, finalshape)

        step = 0

        result = self.process_basic(
            images_info,
            step=step,
            target_is_sky=target_is_sky,
            extinction=extinction,
            method=method,
            method_kwargs=method_kwargs
        )

        if rinput.refine_offsets:
            self.logger.debug("Compute cross-correlation of images")
            # regions_c = self.compute_regions(finalshape, box=200, corners=True)

            # Regions from bright objects
            regions_c = self.compute_regions_from_objs(
                step, result[0].data, finalshape, box=40
            )

            try:

                offsets_xy_c = self.compute_offset_xy_crosscor_regions(
                    images_info, regions_c, refine=True, tol=1
                )
                #
                # Combined offsets
                # Offsets in numpy order, swaping
                offset_xy0 = numpy.fliplr(offset_fc0)
                offsets_xy_t = offset_xy0 - offsets_xy_c
                offsets_fc = numpy.fliplr(offsets_xy_t)
                offsets_fc_t = numpy.round(offsets_fc).astype('int')
                self.logger.debug('Total offsets:\n%s', offsets_xy_t)
                self.logger.info('Computing relative offsets from cross-corr')
                finalshape2, offsetsp2 = narray.combine_shape(
                    baseshape, offsets_fc_t
                )
                #
                self.logger.debug("Relative offsetsp (crosscorr):\n%s",
                                  offsetsp2)
                self.logger.info('Shape of resized array (crosscorr) is '
                                 '(NAXIS2, NAXIS1) = %s', finalshape2)

                # Resizing target imgs
                self.logger.debug("Resize to final offsets")
                self.resize(target_info, baseshape, offsetsp2, finalshape2)
                result = self.process_basic(
                    images_info,
                    step=step,
                    target_is_sky=target_is_sky,
                    extinction=extinction,
                    method=method,
                    method_kwargs=method_kwargs
                )

            except Exception as error:
                self.logger.warning('Error during cross-correlation, %s',
                                    error)

        step = 1

        while step <= rinput.iterations:
            result = self.process_advanced(
                images_info, result, step, target_is_sky,
                maxsep=sky_images_sep_time, nframes=sky_images,
                extinction=extinction,
                method=method, method_kwargs=method_kwargs,
                nside_adhoc_sky_correction=nside_adhoc_sky_correction
            )
            step += 1

        return self.create_result(result_image=result)

    def compute_offset_xy_crosscor_regions(self, iinfo, regions, refine=False,
                                           tol=0.5):

        names = [frame.lastname for frame in iinfo]
        with manage_fits(names) as imgs:
            arrs = [img[0].data for img in imgs]
            offsets_xy = offsets_from_crosscor_regions(
                arrs, regions,
                refine=refine, order='xy', tol=tol
            )
            self.logger.debug("offsets_xy cross-corr:\n%s", offsets_xy)
        return offsets_xy

    def compute_size(self, images_info, baseshape, user_offsets=None):

        # Reference pixel in the center of the frame
        refpix = numpy.array([[baseshape[0] / 2.0, baseshape[1] / 2.0]])

        target_info = [iinfo for iinfo in images_info if iinfo.valid_target]

        if user_offsets is not None:
            self.logger.info('Using offsets from parameters')
            base_ref = numpy.asarray(user_offsets)
            list_of_offsets = -(base_ref - base_ref[0])
        else:
            self.logger.info('Computing offsets from WCS information')
            with manage_fits(img.origin for img in target_info) as images:
                list_of_offsets = offsets_from_wcs_imgs(images, refpix)

        # FIXME: I am using offsets in row/columns
        # the values are provided in XY so flip-lr
        list_of_offsets = numpy.fliplr(list_of_offsets)

        # Insert pixel offsets between frames
        for iinfo, off in zip(target_info, list_of_offsets):
            # Insert pixel offsets between frames
            iinfo.pix_offset = off

            self.logger.debug('Frame %s, offset=%s',
                              iinfo.label, off)

        self.logger.info('Computing relative offsets')
        offsets = [iinfo.pix_offset for iinfo in target_info]
        offsets = numpy.round(offsets).astype('int')

        finalshape, offsetsp = narray.combine_shape(baseshape, offsets)
        self.logger.debug("Relative offsetsp:\n%s", offsetsp)
        self.logger.info('Shape of resized array is (NAXIS2, NAXIS1) = %s',
                         finalshape)
        return finalshape, offsetsp, refpix, list_of_offsets

    def process_basic(self, images_info, step=None,
                      target_is_sky=True,
                      extinction=0.0,
                      method=None, method_kwargs=None):

        target_info = [iinfo for iinfo in images_info if iinfo.valid_target]
        sky_info = [iinfo for iinfo in images_info if iinfo.valid_sky]

        self.logger.info("Step %d, SF: compute superflat", step)
        sf_arr = self.compute_superflat(
            images_info,
            method=method, method_kwargs=method_kwargs
        )

        # Apply superflat
        self.logger.info("Step %d, SF: apply superflat", step)
        for iinfo in images_info:
            self.correct_superflat(iinfo, sf_arr, step=step, save=True)

        self.logger.info('Simple sky correction')
        if target_is_sky:
            # Each frame is the closest sky frame available
            for iinfo in images_info:
                self.compute_simple_sky_for_frame(iinfo, iinfo)
        else:
            # Not implemented
            self.compute_simple_sky(target_info, sky_info)

        # Combining the frames
        self.logger.info("Step %d, Combining target frames", step)
        result = self.combine_frames(target_info, extinction=extinction,
                                     method=method, method_kwargs=method_kwargs)
        self.logger.info('Step %d, finished', step)

        return result

    def process_advanced(self, images_info, result, step, target_is_sky=True,
                         maxsep=5.0, nframes=6, extinction=0,
                         method=None, method_kwargs=None,
                         nside_adhoc_sky_correction=0):

        seeing_fwhm = None
        baseshape = (EMIR_NAXIS2, EMIR_NAXIS1)
        target_info = [iinfo for iinfo in images_info if iinfo.valid_target]
        sky_info = [iinfo for iinfo in images_info if iinfo.valid_sky]
        self.logger.info('Step %d, generating segmentation image', step)

        objmask, seeing_fwhm = self.create_mask(
            result, seeing_fwhm, step=step)

        for frame in target_info:
            frame.objmask = name_object_mask(frame.label, step)
            self.logger.info(
                'Step %d, create object mask %s', step, frame.objmask)
            frame.objmask_data = objmask[frame.valid_region]
            fits.writeto(
                frame.objmask, frame.objmask_data, overwrite=True)

        if not target_is_sky:
            # Empty object mask for sky frames
            bogus_objmask = numpy.zeros(baseshape, dtype='uint8')

            for frame in sky_info:
                frame.objmask_data = bogus_objmask

        self.logger.info("Step %d, SF: compute superflat", step)
        sf_arr = self.compute_superflat(sky_info, segmask=objmask, step=step,
                                        method=method, method_kwargs=method_kwargs)

        # Apply superflat
        self.logger.info("Step %d, SF: apply superflat", step)
        for iinfo in images_info:
            self.correct_superflat(iinfo, sf_arr, step=step, save=True)

        self.logger.info('Step %d, advanced sky correction (SC)', step)
        self.compute_advanced_sky(
            target_info, objmask,
            skyframes=sky_info,
            target_is_sky=target_is_sky,
            maxsep=maxsep,
            nframes=nframes,
            step=step,
            method=method,
            method_kwargs=method_kwargs,
            nside_adhoc_sky_correction=nside_adhoc_sky_correction
        )

        # Combining the images
        self.logger.info("Step %d, Combining the images", step)
        # FIXME: only for science
        result = self.combine_frames(target_info, extinction, step=step,
                                     method=method, method_kwargs=method_kwargs)
        return result

    def compute_simple_sky_for_frame(self, frame, skyframe, step=0, save=True):
        self.logger.info('Correcting sky in frame.....: %s', frame.lastname)
        self.logger.info('with sky computed from frame: %s', skyframe.lastname)

        if hasattr(skyframe, 'median_sky'):
            sky = skyframe.median_sky
        else:

            with fits.open(skyframe.lastname, mode='readonly') as hdulist:
                data = hdulist['primary'].data
                valid = data[frame.valid_region]

                if skyframe.objmask_data is not None:
                    self.logger.debug('object mask defined')
                    msk = frame.objmask_data
                    sky = numpy.median(valid[msk == 0])
                else:
                    self.logger.debug('object mask empty')
                    sky = numpy.median(valid)

            self.logger.debug('median sky value is %f', sky)
            skyframe.median_sky = sky

        dst = name_skysub_proc(frame.label, step)
        prev = frame.lastname

        if save:
            shutil.copyfile(prev, dst)
        else:
            os.rename(prev, dst)

        frame.lastname = dst

        with fits.open(frame.lastname, mode='update') as hdulist:
            data = hdulist['primary'].data
            valid = data[frame.valid_region]
            valid -= sky
            self.logger.info('Sky-subtrated image in frame: %s',
                             frame.lastname)
            self.logger.info('---')

    def compute_simple_sky(self, frame, skyframe, step=0, save=True):
        raise NotImplementedError

    def correct_superflat(self, frame, fitted, step=0, save=True):

        frame.flat_corrected = name_skyflat_proc(frame.label, step)
        if save:
            shutil.copyfile(frame.resized_base, frame.flat_corrected)
        else:
            os.rename(frame.resized_base, frame.flat_corrected)

        self.logger.info("Step %d, SF: apply superflat to frame %s",
                         step, frame.flat_corrected)
        with fits.open(frame.flat_corrected, mode='update') as hdulist:
            data = hdulist['primary'].data
            datar = data[frame.valid_region]
            data[frame.valid_region] = narray.correct_flatfield(datar, fitted)

            frame.lastname = frame.flat_corrected

    def initial_classification(self, obresult, target_is_sky=False):
        """Classify input frames, """
        # lists of targets and sky frames

        with obresult.frames[0].open() as baseimg:
            # Initial checks
            has_bpm_ext = 'BPM' in baseimg
            self.logger.info('images have BPM extension: %s', has_bpm_ext)

        images_info = []
        for f in obresult.frames:
            with f.open() as img:
                # Getting some metadata from FITS header
                hdr = img[0].header

                iinfo = ImageInfo(f)

                finfo = {}
                iinfo.metadata = finfo

                finfo['uuid'] = hdr['UUID']
                finfo['exposure'] = hdr['EXPTIME']
                # frame.baseshape = get_image_shape(hdr)
                finfo['airmass'] = hdr['airmass']
                finfo['mjd'] = hdr['tstamp']

                iinfo.label = 'result_image_{}'.format(finfo['uuid'])
                iinfo.mask = nfcom.Extension("BPM")
                # Insert pixel offsets between frames
                iinfo.objmask_data = None
                iinfo.valid_target = False
                iinfo.valid_sky = False

                # FIXME: hardcode itype for the moment
                iinfo.itype = 'TARGET'
                if iinfo.itype == 'TARGET':
                    iinfo.valid_target = True
                    # targetframes.append(iinfo)
                    if target_is_sky:
                        iinfo.valid_sky = True
                        # skyframes.append(iinfo)
                if iinfo.itype == 'SKY':
                    iinfo.valid_sky = True
                    # skyframes.append(iinfo)
                images_info.append(iinfo)

        return images_info

    def compute_superflat(self, images_info, segmask=None, step=0,
                          method=None, method_kwargs=None):

        self.logger.info("Step %d, SF: combining the frames without offsets",
                         step)

        base_imgs = [img.resized_base for img in images_info]
        with manage_fits(base_imgs) as imgs:

            data = []
            masks = []

            for img, img_info in zip(imgs, images_info):
                self.logger.debug('Step %d, opening resized frame %s',
                                  step, img_info.resized_base)
                data.append(img['primary'].data[img_info.valid_region])

            scales = [numpy.median(d) for d in data]

            if segmask is not None:
                masks = [segmask[frame.valid_region] for frame in images_info]
            else:
                for frame in images_info:
                    self.logger.debug('Step %d, opening resized mask  %s',
                                      step, frame.resized_mask)
                    hdulist = fits.open(
                         frame.resized_mask, memmap=True, mode='readonly')
                    # filelist.append(hdulist)
                    masks.append(hdulist['primary'].data[frame.valid_region])
                masks = None

            self.logger.debug("Step %d, combining %d frames using '%s'",
                              step, len(data), method.__name__)
            sf_data, _sf_var, sf_num = method(
                data, masks, scales=scales, dtype='float32', **method_kwargs
            )

        # Normalize, flat has mean = 1
        sf_data[sf_data == 0] = 1e-5
        sf_data /= sf_data.mean()
        # sf_data[sf_data <= 0] = 1.0

        # Auxiliary data
        sfhdu = fits.PrimaryHDU(sf_data)
        self.save_intermediate_img(sfhdu, name_skyflat('comb', step))
        return sf_data

    '''
    def compute_sky_advanced(self, data_hdul, omasks, base_header, use_errors):
        method = narray.combine.mean

        self.logger.info("recombine images with segmentation mask using '%s'", method.__name__)
        sky_data = method([m[0].data for m in data_hdul], masks=omasks, dtype='float32')

        hdu = fits.PrimaryHDU(sky_data[0], header=base_header)
        points_no_data = (sky_data[2] == 0).sum()

        self.logger.debug('update created sky image result header')
        skyid = str(uuid.uuid1())
        hdu.header['UUID'] = skyid
        hdu.header['history'] = "Combined {} images using '{}'".format(
            len(data_hdul),
            method.__name__
        )
        hdu.header['history'] = 'Combination time {}'.format(
            datetime.datetime.utcnow().isoformat()
        )
        for img in data_hdul:
            hdu.header['history'] = "Image {}".format(self.datamodel.get_imgid(img))

        msg = "missing pixels, total: {}, fraction: {:3.1f}".format(
            points_no_data,
            points_no_data / sky_data[2].size
        )
        hdu.header['history'] = msg
        self.logger.debug(msg)

        if use_errors:
            varhdu = fits.ImageHDU(sky_data[1], name='VARIANCE')
            num = fits.ImageHDU(sky_data[2], name='MAP')
            sky_result = fits.HDUList([hdu, varhdu, num])
        else:
            sky_result = fits.HDUList([hdu])

        return sky_result
    '''

    def combine_frames(self, frames, extinction, out=None, step=0,
                       method=None, method_kwargs=None):
        self.logger.debug('Step %d, opening sky-subtracted frames', step)

        def fits_open(name):
            """Open FITS with memmap in readonly mode"""
            return fits.open(name, mode='readonly', memmap=True)

        frameslll = [fits_open(frame.lastname)
                     for frame in frames if frame.valid_target]
        self.logger.debug('Step %d, opening mask frames', step)
        mskslll = [fits_open(frame.resized_mask)
                   for frame in frames if frame.valid_target]

        self.logger.debug("Step %d, combining %d frames using '%s'",
                          step, len(frameslll), method.__name__)
        try:
            extinc = [pow(10, -0.4 * frame.metadata['airmass'] * extinction)
                      for frame in frames if frame.valid_target]
            data = [i['primary'].data for i in frameslll]
            masks = [i['primary'].data for i in mskslll]
            headers = [i['primary'].header for i in frameslll]

            out = method(data, masks, scales=extinc, dtype='float32', out=out,
                         **method_kwargs)

            base_header = headers[0]
            hdu = fits.PrimaryHDU(out[0], header=base_header)
            hdu.header['history'] = "Combined %d images using '%s'" % (len(frameslll), method.__name__)
            hdu.header['history'] = 'Combination time {}'.format(datetime.datetime.utcnow().isoformat())
            for img in frameslll:
                hdu.header['history'] = "Image {}".format(img[0].header['uuid'])
            prevnum = base_header.get('NUM-NCOM', 1)
            hdu.header['NUM-NCOM'] = prevnum * len(frameslll)
            hdu.header['NUMRNAM'] = 'FullDitheredImagesRecipe'
            hdu.header['UUID'] = str(uuid.uuid1())
            hdu.header['OBSMODE'] = 'FULL_DITHERED_IMAGE'
            # Headers of last image
            hdu.header['TSUTC2'] = headers[-1]['TSUTC2']

            varhdu = fits.ImageHDU(out[1], name='VARIANCE')
            num = fits.ImageHDU(out[2].astype('uint8'), name='MAP')

            result = fits.HDUList([hdu, varhdu, num])
            # saving the three extensions
            fits.writeto('result_i%0d.fits' % step, out[0], overwrite=True)
            fits.writeto('result_i%0d_var.fits' % step, out[1], overwrite=True)
            fits.writeto('result_i%0d_npix.fits' % step, out[2], overwrite=True)

            result.writeto('result_i%0d_full.fits' % step, overwrite=True)
            return result

        finally:
            self.logger.debug('Step %d, closing sky-subtracted frames', step)
            for f in frameslll:
                f.close()
            self.logger.debug('Step %d, closing mask frames', step)
            for f in mskslll:
                f.close()

    def resize(self, frames, shape, offsetsp, finalshape, window=None,
               scale=1, step=0):
        self.logger.info('Resizing frames and masks')
        self.logger.debug('shape, finalshape (NAXIS2, NAXIS1) = %s --> %s',
                          shape, finalshape)
        for frame, rel_offset in zip(frames, offsetsp):
            if frame.valid_target:
                region, _ = narray.subarray_match(finalshape, rel_offset, shape)
                # Valid region
                frame.valid_region = region
                # Relative offset
                frame.rel_offset = rel_offset
                # names of frame and mask
                framen, maskn = name_redimensioned_frames(
                    frame.label, step)
                frame.resized_base = framen
                frame.resized_mask = maskn
                self.logger.debug('%s', frame.label)
                self.logger.debug('valid region is %s, relative offset is %s',
                                  custom_region_to_str(region), rel_offset)
                self.resize_frame_and_mask(
                    frame, finalshape, framen, maskn, window, scale)
                self.logger.debug('---')

    def resize_frame_and_mask(self, frame, finalshape,
                              framen, maskn, window, scale):
        self.logger.info('Resizing frame %s', frame.label)
        with frame.origin.open() as hdul:
            baseshape = hdul[0].data.shape

            # FIXME: Resize_fits saves the resized image in framen
            resize_fits(hdul, framen, finalshape, frame.valid_region,
                        window=window, scale=scale, dtype='float32')

        self.logger.info('Resizing mask  %s', frame.label)
        # We don't conserve the sum of the values of the frame here, just
        # expand the mask

        if frame.mask is None:
            self.logger.warning('BPM missing, use zeros instead')
            false_mask = numpy.zeros(baseshape, dtype='int16')
            hdum = fits.HDUList(fits.PrimaryHDU(false_mask))
            frame.mask = hdum  # DataFrame(frame=hdum)
        elif isinstance(frame.mask, nfcom.Extension):
            ename = frame.mask.name
            with frame.origin.open() as hdul:
                frame.mask = fits.HDUList(hdul[ename].copy())

        resize_fits(frame.mask, maskn, finalshape, frame.valid_region,
                    fill=1, window=window, scale=scale, conserve=False)

    def create_mask(self, img, seeing_fwhm, step=0):

        #
        remove_border = True

        # sextractor takes care of bad pixels

        # if seeing_fwhm is not None and seeing_fwhm > 0:
        #    sex.config['SEEING_FWHM'] = seeing_fwhm * sex.config['PIXEL_SCALE']

        if remove_border:
            weigthmap = 'weights4rms.fits'

            # Create weight map, remove n pixs from either side
            # using a Hannig filter
            # npix = 90
            # w1 = npix
            # w2 = npix
            # wmap = numpy.ones_like(sf_data[0])

            # cos_win1 = numpy.hanning(2 * w1)
            # cos_win2 = numpy.hanning(2 * w2)

            # wmap[:,:w1] *= cos_win1[:w1]
            # wmap[:,-w1:] *= cos_win1[-w1:]
            # wmap[:w2,:] *= cos_win2[:w2, numpy.newaxis]
            # wmap[-w2:,:] *= cos_win2[-w2:, numpy.newaxis]

            # Take the number of combined images from the combined image
            wm = img[2].data.copy()
            # Dont search objects where nimages < lower
            # FIXME: this is a magic number
            # We ignore objects in regions where we have less
            # than 10% of the images
            lower = wm.max() // 10
            border = (wm < lower)
            fits.writeto(weigthmap, border.astype('uint8'), overwrite=True)

            # sex.config['WEIGHT_TYPE'] = 'MAP_WEIGHT'
            # FIXME: this is a magic number
            # sex.config['WEIGHT_THRESH'] = 50
            # sex.config['WEIGHT_IMAGE'] = weigthmap
        else:
            border = None

        data_res = img[0].data
        bkg = sep.Background(data_res)
        data_sub = data_res - bkg

        self.logger.info('Runing source extraction in previous result')
        objects, objmask = sep.extract(data_sub, 1.5, err=bkg.globalrms,
                                       mask=border, segmentation_map=True)
        fits.writeto(name_segmask(step), objmask, overwrite=True)

        # # Plot objects
        # # FIXME, plot sextractor objects on top of image
        # patches = []
        # fwhms = []
        # nfirst = 0
        # catalog_f = sopen(sex.config['CATALOG_NAME'])
        # try:
        #     star = catalog_f.readline()
        #     while star:
        #         flags = star['FLAGS']
        #         # ignoring those objects with corrupted apertures
        #         if flags & sexcatalog.CORRUPTED_APER:
        #             star = catalog_f.readline()
        #             continue
        #         center = (star['X_IMAGE'], star['Y_IMAGE'])
        #         wd = 10 * star['A_IMAGE']
        #         hd = 10 * star['B_IMAGE']
        #         color = 'red'
        #         e = Ellipse(center, wd, hd, star['THETA_IMAGE'], color=color)
        #         patches.append(e)
        #         fwhms.append(star['FWHM_IMAGE'])
        #         nfirst += 1
        #         # FIXME Plot a ellipse
        #         star = catalog_f.readline()
        # finally:
        #     catalog_f.close()
        #
        # p = PatchCollection(patches, alpha=0.4)
        # ax = self._figure.gca()
        # ax.add_collection(p)
        # self._figure.canvas.draw()
        # self._figure.savefig('figure-segmentation-overlay_%01d.png' % step)
        #
        # self.figure_fwhm_histogram(fwhms, step=step)
        #
        # # mode with an histogram
        # hist, edges = numpy.histogram(fwhms, 50)
        # idx = hist.argmax()
        #
        # seeing_fwhm = 0.5 * (edges[idx] + edges[idx + 1])
        # if seeing_fwhm <= 0:
        #     _logger.warning(
        #         'Seeing FHWM %f pixels is negative, reseting', seeing_fwhm)
        #     seeing_fwhm = None
        # else:
        #     _logger.info('Seeing FHWM %f pixels (%f arcseconds)',
        #                  seeing_fwhm, seeing_fwhm * sex.config['PIXEL_SCALE'])
        # objmask = fits.getdata(name_segmask(step))

        return objmask, seeing_fwhm

    def compute_advanced_sky(self, targetframes, objmask,
                             skyframes=None, target_is_sky=False,
                             maxsep=5.0,
                             nframes=10,
                             step=0,
                             method=None, method_kwargs=None,
                             nside_adhoc_sky_correction=0):

        if target_is_sky:
            skyframes = targetframes
            # Each frame is its closest sky frame
            nframes += 1
        elif skyframes is None:
            raise ValueError('skyframes not defined')

        # build kdtree
        sarray = numpy.array([frame.metadata['mjd'] for frame in skyframes])
        # shape must be (n, 1)
        sarray = numpy.expand_dims(sarray, axis=1)

        # query
        tarray = numpy.array([frame.metadata['mjd'] for frame in targetframes])
        # shape must be (n, 1)
        tarray = numpy.expand_dims(tarray, axis=1)

        kdtree = KDTree(sarray)

        # 1 / minutes in a Julian day
        SCALE = 60.0
        # max_time_sep = ri.sky_images_sep_time / 1440.0
        _dis, idxs = kdtree.query(tarray, k=nframes,
                                  distance_upper_bound=maxsep * SCALE)

        nsky = len(sarray)

        for tid, idss in enumerate(idxs):
            try:
                tf = targetframes[tid]
                self.logger.info("Step %d, SC: computing advanced sky for %s using '%s'",
                                 step, tf.label, method.__name__)
                # filter(lambda x: x < nsky, idss)
                locskyframes = []
                for si in idss:
                    if tid == si:
                        # this sky frame it is the current frame, reject
                        continue
                    if si < nsky:
                        self.logger.debug('Step %d, SC: %s is a sky frame',
                                          step, skyframes[si].label)
                        locskyframes.append(skyframes[si])
                self.compute_advanced_sky_for_frame(
                    tf, locskyframes, step=step,
                    method=method, method_kwargs=method_kwargs,
                    nside_adhoc_sky_correction=nside_adhoc_sky_correction
                )
            except IndexError:
                self.logger.error('No sky image available for frame %s', tf.lastname)
                raise

    def compute_advanced_sky_for_frame(self, frame, skyframes, step=0,
                                       method=None, method_kwargs=None,
                                       nside_adhoc_sky_correction=0):
        self.logger.info('Correcting sky in frame %s', frame.lastname)
        self.logger.info('with sky computed from frames')
        for i in skyframes:
            self.logger.info('%s', i.flat_corrected)

        data = []
        scales = []
        masks = []
        # handle the FITS file to close it finally
        desc = []
        try:
            for i in skyframes:
                filename = i.flat_corrected
                hdulist = fits.open(filename, mode='readonly', memmap=True)

                data.append(hdulist['primary'].data[i.valid_region])
                desc.append(hdulist)
                scales.append(numpy.median(data[-1]))
                if i.objmask_data is not None:
                    masks.append(i.objmask_data)
                    self.logger.debug('object mask is shared')
                elif i.objmask is not None:
                    hdulistmask = fits.open(
                        i.objmask, mode='readonly', memmap=True)
                    masks.append(hdulistmask['primary'].data)
                    desc.append(hdulistmask)
                    self.logger.debug('object mask is particular')
                else:
                    self.logger.warn('no object mask for %s', filename)

            self.logger.debug("computing background with %d frames using '%s'",
                              len(data), method.__name__)
            sky, _, num = method(data, masks, scales=scales, **method_kwargs)

            with fits.open(frame.lastname) as hdulist:
                data = hdulist['primary'].data
                valid = data[frame.valid_region]

                if frame.objmask_data is not None:
                    self.logger.debug('object mask defined')
                    msk = frame.objmask_data
                    skymedian = numpy.median(valid[msk == 0])
                else:
                    self.logger.debug('object mask empty')
                    skymedian = numpy.median(valid)
                self.logger.debug('scaling with skymedian %s', skymedian)
            sky *= skymedian

        finally:
            # Closing all FITS files
            for hdl in desc:
                hdl.close()

        if numpy.any(num == 0):
            # We have pixels without
            # sky background information
            self.logger.warning('pixels without sky information when correcting %s',
                                frame.flat_corrected)
            binmask = num == 0
            # FIXME: during development, this is faster
            # sky[binmask] = sky[num != 0].mean()

            # To continue we interpolate over the patches
            narray.fixpix2(sky, binmask, out=sky, iterations=1)

            name = name_skybackgroundmask(frame.label, step)
            fits.writeto(name, binmask.astype('int16'), overwrite=True)

        name_sky = name_skybackground(frame.label, step)
        fits.writeto(name_sky, sky, overwrite=True)

        dst = name_skysub_proc(frame.label, step)
        prev = frame.lastname
        shutil.copyfile(prev, dst)
        frame.lastname = dst

        with fits.open(frame.lastname, mode='update') as hdulist:
            data = hdulist['primary'].data
            valid = data[frame.valid_region]
            valid -= sky
            if nside_adhoc_sky_correction > 0:
                skycorr = self.adhoc_sky_correction(
                    arr=valid,
                    objmask=frame.objmask_data,
                    nside=nside_adhoc_sky_correction
                )
                valid -= skycorr

    def compute_regions_from_objs(self, step, arr, finalshape, box=50,
                                  corners=True):
        regions = []
        # create catalog of objects skipping a border around the image
        catalog, mask = self.create_object_catalog(arr, border=300)

        self.save_intermediate_array(mask, 'objmask_i{}.fits'.format(step))
        # with the catalog, compute the brightest NKEEP objects

        LIMIT_AREA = 5000
        NKEEP = 3
        idx_small = catalog['npix'] < LIMIT_AREA
        objects_small = catalog[idx_small]
        idx_flux = objects_small['flux'].argsort()
        objects_nth = objects_small[idx_flux][-NKEEP:]
        for obj in objects_nth:
            self.logger.debug('ref is (x,y) = (%s, %s)', obj['x'], obj['y'])
            region = nautils.image_box2d(obj['x'], obj['y'], finalshape, (box, box))
            regions.append(region)
        return regions

    def create_object_catalog(self, arr, threshold=3.0, border=0):

        if border > 0:
            wmap = numpy.ones_like(arr)
            wmap[border:-border, border:-border] = 0
        else:
            wmap = None

        bkg = sep.Background(arr)
        data_sub = arr - bkg
        objects, objmask = sep.extract(
            data_sub,
            threshold,
            err=bkg.globalrms * numpy.ones_like(data_sub),
            mask=wmap,
            segmentation_map=True
        )
        return objects, objmask

    def adhoc_sky_correction(self, arr, objmask, nside=10):
        # nside: number of subdivisions in each quadrant

        self.logger.info('computing ad hoc sky correction')

        skyfit = numpy.zeros_like(arr)

        # fit each quadrant separately
        lim = [0, 1024, 2048]
        for i in range(2):
            i1 = lim[i]
            i2 = lim[i + 1]
            for j in range(2):
                j1 = lim[j]
                j2 = lim[j + 1]
                xyfit = []
                zfit = []
                limi = numpy.linspace(i1, i2, nside + 1, dtype=int)
                limj = numpy.linspace(j1, j2, nside + 1, dtype=int)
                for ii in range(nside):
                    ii1 = limi[ii]
                    ii2 = limi[ii+1]
                    for jj in range(nside):
                        jj1 = limj[jj]
                        jj2 = limj[jj+1]
                        tmprect = arr[ii1:ii2, jj1:jj2]
                        tmpmask = objmask[ii1:ii2, jj1:jj2]
                        usefulpix = tmprect[tmpmask == 0]
                        if usefulpix.size > 0:
                            x0 = (jj1 + jj2) / 2.0
                            y0 = (ii1 + ii2) / 2.0
                            z0 = numpy.median(usefulpix)
                            xyfit.append([x0, y0])
                            zfit.append(z0)
                xyfit = numpy.array(xyfit)
                zfit = numpy.array(zfit)
                xgrid, ygrid = numpy.meshgrid(
                    numpy.arange(j1, j2, dtype=float),
                    numpy.arange(i1, i2, dtype=float))
                surface_nearest = interpolate.griddata(
                    xyfit, zfit, (xgrid, ygrid), method='nearest',
                    rescale=True
                )
                surface_cubic = interpolate.griddata(
                    xyfit, zfit, (xgrid, ygrid), method='cubic',
                    fill_value=-1.0E30,
                    rescale=True
                )
                skyfit[i1:i2, j1:j2] = numpy.where(
                    surface_cubic < -1.0E29,
                    surface_nearest,
                    surface_cubic
                )

        # hdu = fits.PrimaryHDU(arr.astype('float32'))
        # hdul = fits.HDUList([hdu])
        # hdul.writeto('xxx1.fits', overwrite=True)
        # hdu = fits.PrimaryHDU(skyfit.astype('float32'))
        # hdul = fits.HDUList([hdu])
        # hdul.writeto('xxx2.fits', overwrite=True)

        return skyfit



#
# Copyright 2015 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

'''Test the procedures module'''

import pytest

import numpy
from numpy.testing import assert_allclose

from ..procedures import encloses_annulus


def test_encloses_annulus():
    a = numpy.zeros((100, 100))
    r_in = 6
    r_out = 15.5
    xc = 51.4
    yc = 52.9

    x_min = -0.5 - xc
    x_max = a.shape[1] - 0.5 - xc
    y_min = -0.5 - yc
    y_max = a.shape[1] - 0.5 - yc


    aa = encloses_annulus(x_min, x_max, y_min, y_max,
                          a.shape[1], a.shape[0],
                          r_in, r_out)

    assert_allclose(aa[50, [20, 40, 50, 60]], [0.0, 1.0, 0.0, 1.0])



#
# Copyright 2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Generate YAML files to reduce ABBA observations"""

import argparse
import sys
import textwrap

from numina.array.display.fileinfo import FileInfo
from numina.tools.arg_file_is_new import arg_file_is_new


# auxiliary function to generate content of YAML files
def generate_yaml_content(args, list_fileinfo, enabled=True):

    step = args.step

    # obsid_prefix
    if args.obsid_prefix is None:
        obsid_prefix = ''
    else:
        obsid_prefix = args.obsid_prefix + '_'

    nimages = len(list_fileinfo)

    output = ''
    if step == 0:
        # initial rectification and wavelength calibration coefficients
        idlabel = list_fileinfo[0].filename[:10]
        if obsid_prefix == '':
            output += 'id: _{}_rectwv_preliminary\n'.format(idlabel)
        else:
            output += 'id: _{}_{}_rectwv_preliminary\n'.format(obsid_prefix,
                                                               idlabel)
        output += 'instrument: EMIR\n'
        output += 'mode: GENERATE_RECTWV_COEFF\n'
        output += 'frames:\n'
        for i in range(args.npreliminary):
            output += ' - ' + list_fileinfo[i].filename + '\n'
        if enabled:
            output += 'enabled: True'
        else:
            output += 'enabled: False'
    elif step == 1:
        # check that all the expected parameters have been provided
        lrequirements = [
            'refine_wavecalib_mode',
            'minimum_slitlet_width_mm',
            'maximum_slitlet_width_mm',
            'global_integer_offset_x_pix',
            'global_integer_offset_y_pix'
        ]
        for item in lrequirements:
            if args.__dict__[item] is None:
                raise ValueError('Parameter {} is None!'.format(item))
        # refined rectification and wavelength calibration for each block
        for i in range(nimages):
            if args.rectwv_combined:
                if i == 0:
                    display_id = True
                else:
                    display_id = False
            else:
                display_id = True
            if display_id:
                idlabel = list_fileinfo[i].filename[:10]
                if obsid_prefix == '':
                    output += 'id: _{}_rectwv'.format(idlabel)
                else:
                    output += 'id: _{}{}_rectwv'.format(obsid_prefix, idlabel)
                if args.rectwv_combined:
                    output += '_combined'
                output += '\n'
                output += 'instrument: EMIR\n'
                output += 'mode: GENERATE_RECTWV_COEFF\n'
                output += 'frames:\n'
            output += ' - ' + list_fileinfo[i].filename + '\n'
            if args.rectwv_combined:
                if i == nimages - 1:
                    display_requirements = True
                else:
                    display_requirements = False
            else:
                display_requirements = True
            if display_requirements:
                output += 'requirements:\n'
                for item in lrequirements:
                    output += '  {}: {}\n'.format(item, args.__dict__[item])
                if enabled:
                    output += 'enabled: True'
                else:
                    output += 'enabled: False'
                if i < nimages - 1:
                    output += '\n---\n'
    elif step == 2:
        # apply rectification and wavelength calibration to each block
        if obsid_prefix == '':
            output += 'id: _abba_fast\n'
        else:
            output += 'id: _' + obsid_prefix + '_abba_fast\n'
        output += 'instrument: EMIR\n'
        output += 'mode: ABBA_SPECTRA_FAST_RECTWV\n'
        output += 'frames:\n'
        for i in range(nimages):
            output += ' - ' + list_fileinfo[i].filename + '\n'
        output += 'requirements:\n'
        output += '  pattern: ABBA\n'
        idlabel = list_fileinfo[0].filename[:10]
        output += '  rectwv_coeff: ../obsid_' + idlabel + '_rectwv_'
        if args.rectwv_combined:
            output += 'combined_'
        output += 'results/rectwv_coeff.json\n'
        output += '  method: sigmaclip\n'
        output += '  method_kwargs:\n'
        output += '    low: 3.0\n'
        output += '    high: 3.0\n'
        output += '  voffset_pix: 0.0\n'
        output += 'enabled: True\n'
    elif step == 3:
        # apply rectification and wavelength calibration to each block
        if obsid_prefix == '':
            output += 'id: _abba\n'
        else:
            output += 'id: _' + obsid_prefix + '_abba\n'
        output += 'instrument: EMIR\n'
        output += 'mode: ABBA_SPECTRA_RECTWV\n'
        output += 'frames:\n'
        for i in range(nimages):
            output += ' - ' + list_fileinfo[i].filename + '\n'
        output += 'requirements:\n'
        output += '  pattern: ABBA\n'
        if args.rectwv_combined:
            output += '  rectwv_coeff: ../obsid_'
            output += list_fileinfo[0].filename[:10]
            output += '_rectwv_combined_results/rectwv_coeff.json\n'
        else:
            output += '  list_rectwv_coeff:\n'
            for i in range(nimages):
                idlabel = list_fileinfo[i].filename[:10]
                output += '    - ../obsid_' + idlabel + \
                          '_rectwv_results/rectwv_coeff.json\n'
        output += '  method: sigmaclip\n'
        output += '  method_kwargs:\n'
        output += '    low: 3.0\n'
        output += '    high: 3.0\n'
        output += '  refine_target_along_slitlet:\n'
        output += '    npix_removed_near_ohlines: 3\n'
        output += '    nwidth_medfilt: 11\n'
        output += '    save_individual_images: 0\n'
        output += '    ab_different_target: TBD\n'
        output += '    vpix_region_a_target: [TBD, TBD]\n'
        output += '    vpix_region_a_sky: [TBD, TBD]\n'
        output += '    vpix_region_b_target: [TBD, TBD]\n'
        output += '    vpix_region_b_sky: [TBD, TBD]\n'
        output += '    list_valid_wvregions_a: [ [TBD, TBD], [TBD, TBD] ]\n'
        output += '    list_valid_wvregions_b: [ [TBD, TBD], [TBD, TBD] ]\n'
        output += 'enabled: True\n'
    else:
        raise ValueError('Unexpected step={}'.format(step))

    return output


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: display arrangement of EMIR CSU bars',
        formatter_class=argparse.RawTextHelpFormatter
    )

    # positional arguments
    parser.add_argument("filename",
                        help="TXT file with list of ABBA FITS files",
                        type=argparse.FileType('rt'))
    parser.add_argument("--step", required=True,
                        help=textwrap.dedent("""\
                        0: preliminary rectwv_coeff.json
                        1: refined rectwv_coeff.json
                        2: ABBA fast reduction
                        3: ABBA careful reduction"""),
                        type=int, choices=[0, 1, 2, 3])
    parser.add_argument("--outfile", required=True,
                        help="Output YAML file name",
                        type=lambda x: arg_file_is_new(parser, x))

    # optional arguments
    parser.add_argument("--pattern",
                        help="Observation pattern",
                        default='ABBA',
                        choices=['A', 'AB', 'ABBA'])
    parser.add_argument("--repeat",
                        help="Repetitions at each position",
                        default=1, type=int)
    parser.add_argument("--npreliminary",
                        help="number of images to be combined to compute "
                             "preliminary rectwv_coeff.json",
                        type=int, default=1)
    parser.add_argument("--refine_wavecalib_mode",
                        help=textwrap.dedent("""\
                        0: no refinement
                        1: global offset to all the slitlets (ARC lines)
                        2: individual offset to each slitlet (ARC lines)
                        11: global offset to all the slitlets (OH lines)
                        12: individual offset to each slitlet (OH lines)"""),
                        type=int, choices=[0, 1, 2, 11, 12])
    parser.add_argument("--minimum_slitlet_width_mm",
                        type=float)
    parser.add_argument("--maximum_slitlet_width_mm",
                        type=float)
    parser.add_argument("--global_integer_offset_x_pix",
                        type=int)
    parser.add_argument("--global_integer_offset_y_pix",
                        type=int)
    parser.add_argument("--obsid_prefix",
                        type=str)
    parser.add_argument("--rectwv_combined",
                        help="Generate single rectwv_coeff.json",
                        action="store_true")
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # read TXT file
    with args.filename as f:
        file_content = f.read().splitlines()
    list_fileinfo = []
    for line in file_content:
        if len(line) > 0:
            if line[0] not in ['#', '@']:
                tmplist = line.split()
                tmpfile = tmplist[0]
                if len(tmplist) > 1:
                    tmpinfo = tmplist[1:]
                else:
                    tmpinfo = None
                list_fileinfo.append(FileInfo(tmpfile, tmpinfo))

    # check consistency of pattern, repeat and number of images
    nimages = len(list_fileinfo)
    pattern_sequence = ''
    for i in range(len(args.pattern)):
        pattern_sequence += args.pattern[i] * args.repeat
    if nimages % len(pattern_sequence) != 0:
        raise ValueError('Unexpected number of images')
    nsequences = nimages // len(pattern_sequence)
    full_set = pattern_sequence * nsequences

    print('Expected sequence pattern: {}'.format(pattern_sequence))
    print('Number of sequences......: {}'.format(nsequences))
    print('Full set of images.......: {}'.format(full_set))

    output = generate_yaml_content(args, list_fileinfo)

    # generate YAML file
    with args.outfile as f:
        f.write(output)
    print('--> File {} generated!'.format(args.outfile.name))


if __name__ == "__main__":
    main()



import pytest
import astropy.io.fits as fits
import numpy

import numina.core
import numina.exceptions

from ..subs import BaseABBARecipe
from ..coadd import CoaddABBARecipe

def create_frame(val=0, pos=3, keys=None):
    data = numpy.zeros((10, 10))
    data[pos] += val
    hdu = fits.PrimaryHDU(data)
    if keys:
        for k, v in keys.items():
            hdu.header[k] = v

    return fits.HDUList([hdu])


def create_frame_dark():
    data = numpy.zeros((10, 10))
    hdu = fits.PrimaryHDU(data)
    hdu.header['UUID'] = '11111111111111111111111111111111'
    return fits.HDUList([hdu])


def create_frame_flat():
    data = numpy.ones((10, 10))
    hdu = fits.PrimaryHDU(data)
    hdu.header['UUID'] = '2222222222222222222222222222222'
    return fits.HDUList([hdu])


@pytest.mark.parametrize("nabba", [1,2,3,4])
def test_coadd(nabba):

    expected_data = numpy.zeros((10, 10))
    expected_data[5] = 4
    expected_data[3] = -4

    results = []
    starttime = 1000000001.00034
    for i in range(nabba):
        pos = [5,3,3,5]
        frames = []
        for j in range(4):
            base = starttime + i * 1000
            off1 = j * 100
            off2 = off1 + 100
            t1 = base + off1
            t2 = base + off2
            keys = {'TSUTC1': t1, 'TSUTC2': t2}
            frame = numina.core.DataFrame(frame=create_frame(val=2, pos=pos[j], keys=keys))
            frames.append(frame)

        obsresult = numina.core.ObservationResult()
        obsresult.frames = frames

        recipe = BaseABBARecipe()
        rinput = recipe.create_input(
            obresult=obsresult,
            master_dark=numina.core.DataFrame(frame=create_frame_dark()),
            master_flat=numina.core.DataFrame(frame=create_frame_flat())
        )
        result = recipe.run(rinput)
        results.append(result)

    obsresult = numina.core.ObservationResult()
    obsresult.frames = [result.reduced_mos_abba for result in results]

    recipe = CoaddABBARecipe()
    rinput = recipe.create_input(
        obresult=obsresult,
    )

    result = recipe.run(rinput)
    hdu = result.reduced_mos_abba.frame[0]
    assert hdu.header['NUM-NCOM'] == nabba * 2
    assert hdu.header['TSUTC1'] == starttime
    assert hdu.header['TSUTC2'] == starttime + (nabba - 1) * 1000 + 4 * 100
    assert numpy.allclose(expected_data, hdu.data)


def test_coadd_empty():

    obsresult = numina.core.ObservationResult()
    obsresult.frames = []

    recipe = CoaddABBARecipe()
    rinput = recipe.create_input(
        obresult=obsresult,
    )

    with pytest.raises(numina.exceptions.RecipeError):
        recipe.run(rinput)



#
# Copyright 2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
import pkgutil
from six import StringIO


def retrieve_catlines(mode, grism_name):
    """Retrieve arc/OH lines

    Parameters
    ----------
    mode : int
        Integer, indicating the type of lines:
        1 or 2 : arc lines
        11 or 12 : OH lines
    grism_name : string
        Grism name.

    Returns
    -------
    catlines_all_wave : numpy array
        Array with wavelengths
    catlines_all_flux : numpy array
        Array with fluxes.

    """

    # protections
    if mode not in [1, 2, 11, 12]:
        raise ValueError('Invalid mode={}'.format(mode))

    # read tabulated lines
    if mode in [1, 2]:  # ARC lines
        if grism_name == 'LR':
            catlines_file = 'lines_argon_neon_xenon_empirical_LR.dat'
        else:
            catlines_file = 'lines_argon_neon_xenon_empirical.dat'
        dumdata = pkgutil.get_data('emirdrp.instrument.configs', catlines_file)
        arc_lines_tmpfile = StringIO(dumdata.decode('utf8'))
        catlines = np.genfromtxt(arc_lines_tmpfile)
        # define wavelength and flux as separate arrays
        catlines_all_wave = catlines[:, 0]
        catlines_all_flux = catlines[:, 1]
    elif mode in [11, 12]:  # OH lines
        dumdata = pkgutil.get_data(
            'emirdrp.instrument.configs',
            'Oliva_etal_2013.dat'
        )
        oh_lines_tmpfile = StringIO(dumdata.decode('utf8'))
        catlines = np.genfromtxt(oh_lines_tmpfile)
        # define wavelength and flux as separate arrays
        catlines_all_wave = np.concatenate((catlines[:, 1], catlines[:, 0]))
        catlines_all_flux = np.concatenate((catlines[:, 2], catlines[:, 2]))
    else:
        raise ValueError('Unexpected mode={}'.format(mode))

    return catlines_all_wave, catlines_all_flux



#
# Copyright 2008-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from __future__ import division


import sys
import math
import logging
import itertools

import numpy as np
import matplotlib.pyplot as plt
import sep
import astropy.units as u
import astropy.wcs
import skimage.filters as filt
import skimage.feature as feat
from scipy import ndimage as ndi
from numina.array.utils import coor_to_pix_1d
from numina.array.bbox import BoundingBox
from numina.array.offrot import fit_offset_and_rotation
from numina.array.peaks.peakdet import refine_peaks
from numina.core import Requirement, Result
from numina.types.qc import QC
import numina.types.array as tarray
from numina.core.requirements import ObservationResultRequirement
import numina.core.query as qmod

import emirdrp.instrument.constants as cons
from emirdrp.core.recipe import EmirRecipe
from emirdrp.core.utils import create_rot2d
from emirdrp.processing.combine import basic_processing_, combine_images
from emirdrp.processing.combine import process_ab, process_abba
import emirdrp.requirements as reqs
import emirdrp.products as prods
import emirdrp.instrument.distortions as dist
from emirdrp.instrument.csuconf import TargetType
import emirdrp.instrument.csuconf as csuconf


def comp_centroid(data, bounding_box, debug_plot=False, plot_reference=None, logger=None):
    """Detect objects in a region and return the centroid of the brightest one"""

    from matplotlib.patches import Ellipse

    if logger is None:
        logger = logging.getLogger(__name__)

    region = bounding_box.slice
    ref_x = region[1].start
    ref_y = region[0].start
    logger.debug('region ofset is %s, %s', ref_x, ref_y)
    subimage = data[region].copy()
    bkg = sep.Background(subimage)
    data_sub = subimage - bkg
    objects = sep.extract(data_sub, 1.5, err=bkg.globalrms)
    # Select brightest object
    logger.debug('%d object found', len(objects))

    if len(objects) == 0:
        # print('No objects')
        return None

    iadx = objects['flux'].argmax()
    # plot background-subtracted image
    maxflux = objects[iadx]

    if debug_plot:
        fig, ax = plt.subplots()
        m, s = np.mean(data_sub), np.std(data_sub)
        ax.imshow(data_sub, interpolation='nearest', cmap='gray',
                  vmin=m - s, vmax=m + s, origin='lower',
                  extent=bounding_box.extent)
        if plot_reference:
            e = Ellipse(xy=(plot_reference[0], plot_reference[1]),
                        width=6,
                        height=6,
                        angle=0)
            e.set_facecolor('none')
            e.set_edgecolor('green')
            ax.add_artist(e)

        # plot an ellipse for each object
        for idx, obj in enumerate(objects):
            e = Ellipse(xy=(obj['x'] + ref_x, obj['y'] + ref_y),
                        width=6 * obj['a'],
                        height=6 * obj['b'],
                        angle=obj['theta'] * 180. / np.pi)
            e.set_facecolor('none')
            if idx == iadx:
                e.set_edgecolor('blue')
            else:
                e.set_edgecolor('red')
            ax.add_artist(e)
        return maxflux['x'], maxflux['y'], ax
    else:
        return maxflux['x'], maxflux['y']


class MaskCheckRecipe(EmirRecipe):
    """
    Acquire a target.

    Recipe for the processing of multi-slit/long-slit check images.

    **Observing modes:**

        * MSM and LSM check

    """

    # Recipe Requirements
    #
    obresult = ObservationResultRequirement(
        query_opts=qmod.ResultOf(
            'STARE_IMAGE.result_image',
            node='children',
            id_field="resultsIds"
        )
    )
    master_bpm = reqs.MasterBadPixelMaskRequirement()

    bars_nominal_positions = Requirement(
        prods.NominalPositions,
        'Nominal positions of the bars'
    )

    # Recipe Products
    slit_image = Result(prods.ProcessedImage)
    object_image = Result(prods.ProcessedImage)
    offset = Result(tarray.ArrayType)
    angle = Result(float)
    matched_slits = Result(tarray.ArrayType, description="Centroids measured in the detector")
    centroids = Result(tarray.ArrayType, description="Centroids measured in the detector")

    def run(self, rinput):
        self.logger.info('starting processing for image acquisition')
        # Combine and masking
        flow = self.init_filters(rinput)

        # count frames
        frames = rinput.obresult.frames

        nframes = len(frames)
        if nframes not in [1, 2, 4]:
            raise ValueError("expected 1, 2 or 4 frames, got {}".format(nframes))

        interm = basic_processing_(frames, flow)

        if nframes == 1:
            hdulist_slit = combine_images(interm[:])
            hdulist_object = hdulist_slit
            # background_subs = False
        elif nframes == 2:
            hdulist_slit = combine_images(interm[0:])
            hdulist_object = process_ab(interm)
            # background_subs = True
        elif nframes == 4:
            hdulist_slit = combine_images(interm[0::3])
            hdulist_object = process_abba(interm)
            # background_subs = True
        else:
            raise ValueError("expected 1, 2 or 4 frames, got {}".format(nframes))

        self.set_base_headers(hdulist_slit[0].header)
        self.set_base_headers(hdulist_object[0].header)

        self.save_intermediate_img(hdulist_slit, 'slit_image.fits')

        self.save_intermediate_img(hdulist_object, 'object_image.fits')

        # Get slits
        # Rotation around (0,0)
        # For other axis, offset is changed
        # (Off - raxis) = Rot * (Offnew - raxis)
        wcs = astropy.wcs.WCS(hdulist_slit[0].header)
        rotaxis = wcs.wcs.crpix - 1
        ipa_deg = hdulist_object[0].header.get("IPA", cons.EMIR_REF_IPA) * u.deg
        rotoff_deg = hdulist_object[0].header.get("ROTOFF", cons.EMIR_REF_IPA) * u.deg

        self.logger.debug('center of rotation (from CRPIX) is %s', wcs.wcs.crpix)

        csu_conf = self.load_csu_conf(hdulist_slit, rinput.bars_nominal_positions)

        # IF CSU is completely open OR there are no references,
        # this is not needed
        if not csu_conf.is_open():
            #self.logger.info('CSU is configured, detecting slits')
            # slits_bb = self.compute_slits(hdulist_slit, csu_conf)

            # Not detecting slits. We trust the model
            slits_bb = None
            image_sep = hdulist_object[0].data.astype('float32')

            self.logger.debug('center of rotation (from CRPIX) is %s', wcs.wcs.crpix)
            self.logger.debug('create adapted WCS from header WCS')
            wcsa = dist.adapt_wcs(wcs, ipa_deg=ipa_deg.value, rotang_deg=rotoff_deg.value)
            self.logger.debug('CD matrix of WCS %s', wcs.wcs.cd)
            self.logger.debug('CD matrix of adapted WCS %s', wcsa.wcs.cd)
            # angle in deg, offset in pixels on the CSU
            offset, angle, qc, coords = compute_off_rotation(
                image_sep, wcsa, csu_conf, slits_bb,
                rotaxis=rotaxis, logger=self.logger,
                debug_plot=False, intermediate_results=self.intermediate_results
            )
        else:
            self.logger.info('CSU is open, not detecting slits')
            offset = [0.0, 0.0]
            angle = 0.0
            coords = ([], [], [], [], [])
            qc = QC.GOOD

        pixsize_na = cons.EMIR_PIXSCALE / cons.GTC_NASMYTH_A_PLATESCALE
        self.logger.debug('Pixel size in NASMYTH_A %s mm', pixsize_na)
        # Offset is returned without units
        o_mm = ((offset * u.pixel) * pixsize_na).to(u.mm)
        ipa_rot = create_rot2d(ipa_deg.to('', equivalencies=u.dimensionless_angles()))
        rotoff_rot = create_rot2d(rotoff_deg.to('', equivalencies=u.dimensionless_angles()))
        self.logger.info('IPA is %s deg', ipa_deg.value)
        self.logger.info('ROTOFF is %s deg', rotoff_deg.value)
        self.logger.info('OFF (mm) %s', o_mm)
        o_mm_ipa = np.dot(ipa_rot, o_mm)
        o_mm_rotoff = np.dot(rotoff_rot, o_mm)

        self.logger.info('=========================================')
        self.logger.info('(WITH IPA) Offset Target in Focal Plane Frame %s mm', o_mm_ipa)
        self.logger.info('Offset Target in Focal Plane Frame %s mm', o_mm_rotoff)
        self.logger.info('=========================================')

        # P2 Pos REF VIRT
        # P1 Pos REF DET
        # Q1 Pos Measured DET
        # Q2 Pos Measured VIRT

        matched_slits, p_d, p_v, q_d, q_v = coords
        centroids_det = np.hstack([p_d, q_d])
        centroids_virt = np.hstack([p_v, q_v])

        # Value in mm
        offset_out = o_mm_rotoff.to(u.mm).value
        # Convert DEG to RAD
        angle_out = angle
        result = self.create_result(
            slit_image=hdulist_slit,
            object_image=hdulist_object,
            offset=offset_out, # offset is in milimeters
            angle=angle_out, # angle is in degrees
            qc=qc,
            matched_slits=matched_slits,
            centroids=centroids_det
        )
        self.logger.info('end processing for image acquisition')
        return result

    def load_csu_conf(self, hdulist, bars_nominal_positions):
        from emirdrp.instrument.dtuconf import DtuConf
        # Get slits
        # hdr = hdulist[0].header
        # Extract DTU and CSU information from headers

        dtuconf = DtuConf.from_img(hdulist)

        # coordinates transformation from DTU coordinates
        # to image coordinates
        # Y inverted
        # XY switched
        # trans1 = [[1, 0, 0], [0,-1, 0], [0,0,1]]
        # trans2 = [[0,1,0], [1,0,0], [0,0,1]]
        trans3 = [[0, -1, 0], [1, 0, 0], [0, 0, 1]]  # T3 = T2 * T1

        vec = np.dot(trans3, dtuconf.coor_r) * u.micron/ cons.EMIR_PIXSIZE
        self.logger.debug('DTU shift is %s', vec)

        self.logger.debug('create bar model')
        barmodel = csuconf.create_bar_models(bars_nominal_positions)

        csu_conf = csuconf.CSUConf(barmodel)
        csu_conf.set_state_from_img(hdulist)

        self.logger.info('loaded CSU named: %s', csu_conf.conf_f)
        if self.intermediate_results:
            # FIXME: coordinates are in VIRT pixels
            self.logger.debug('create bar mask from predictions')
            mask = np.ones_like(hdulist[0].data)
            for i in itertools.chain(csu_conf.LBARS, csu_conf.RBARS):
                bar = csu_conf.bars[i]
                mask[bar.bbox().slice] = 0
            self.save_intermediate_array(mask, 'mask_bars.fits')

            self.logger.debug('create slit mask from predictions')
            mask = np.zeros_like(hdulist[0].data)
            for slit in csu_conf.slits.values():
                mask[slit.bbox().slice] = slit.idx
            self.save_intermediate_array(mask, 'mask_slit.fits')

            self.logger.debug('create slit reference mask from predictions')
            mask1 = np.zeros_like(hdulist[0].data)
            for slit in csu_conf.slits.values():
                if slit.target_type == TargetType.REFERENCE:
                    mask1[slit.bbox().slice] = slit.idx
            self.save_intermediate_array(mask1, 'mask_slit_ref.fits')

        return csu_conf

    def compute_slits(self, hdulist, csu_conf):

        self.logger.debug('finding borders of slits')
        self.logger.debug('not strictly necessary...')
        data = hdulist[0].data
        wcs = astropy.wcs.WCS(hdulist[0].header)
        self.logger.debug('dtype of data %s', data.dtype)

        self.logger.debug('median filter (3x3)')
        image_base = ndi.filters.median_filter(data, size=3)

        # Cast as original type for skimage
        self.logger.debug('casting image to unit16 (for skimage)')
        iuint16 = np.iinfo(np.uint16)
        image = np.clip(image_base, iuint16.min, iuint16.max).astype(np.uint16)

        self.logger.debug('compute Sobel filter')
        # FIXME: compute sob and sob_v is redundant
        sob = filt.sobel(image)
        self.save_intermediate_array(sob, 'sobel_image.fits')
        sob_v = filt.sobel_v(image)
        self.save_intermediate_array(sob_v, 'sobel_v_image.fits')

        # Compute detector coordinates of bars
        all_coords_virt = np.empty((110, 2))
        all_coords_real = np.empty((110, 2))

        # Origin of coordinates is 1
        for bar in csu_conf.bars.values():
            all_coords_virt[bar.idx - 1] = bar.xpos, bar.y0

        # Origin of coordinates is 1 for this function
        _x, _y = dist.wcs_exvp(wcs, all_coords_virt[:, 0], all_coords_virt[:, 1])
        all_coords_real[:, 0] = _x
        all_coords_real[:, 1] = _y

        # FIXME: hardcoded value
        # use what is defined in CSUBarModel instead
        h = 16
        slit_h_virt = 16.242
        slit_h_tol = 3
        slits_bb = {}

        mask1 = np.zeros_like(hdulist[0].data)

        for idx in range(csuconf.EMIR_NBARS):
            lbarid = idx + 1
            rbarid = lbarid + csuconf.EMIR_NBARS
            ref_x_l_v, ref_y_l_v = all_coords_virt[lbarid - 1]
            ref_x_r_v, ref_y_r_v = all_coords_virt[rbarid - 1]

            ref_x_l_d, ref_y_l_d = all_coords_real[lbarid - 1]
            ref_x_r_d, ref_y_r_d = all_coords_real[rbarid - 1]

            width_v = ref_x_r_v - ref_x_l_v
            # width_d = ref_x_r_d - ref_x_l_d

            if (ref_y_l_d >= 2047 + h) or (ref_y_l_d <= 1 - h):
                # print('reference y position is outlimits, skipping')
                continue

            if width_v < 5:
                # print('width is less than 5 pixels, skipping')
                continue

            plot = False
            regionw = 12
            px1 = coor_to_pix_1d(ref_x_l_d) - 1
            px2 = coor_to_pix_1d(ref_x_r_d) - 1
            prow = coor_to_pix_1d(ref_y_l_d) - 1

            comp_l, comp_r = calc_bars_borders(image, sob_v, prow, px1, px2, regionw, h=h,
                                               plot=plot, lbarid=lbarid, rbarid=rbarid,
                                               plot2=False)
            if np.any(np.isnan([comp_l, comp_r])):
                self.logger.warning("converting NaN value, border of=%d", idx + 1)
                self.logger.warning("skipping bar=%d", idx + 1)
                continue
            elif comp_l > comp_r:
                # Not refining
                self.logger.warning("computed left border of=%d greater than right border", idx + 1)
                comp2_l, comp2_r = px1, px2
            else:
                region2 = 5
                px21 = coor_to_pix_1d(comp_l)
                px22 = coor_to_pix_1d(comp_r)

                comp2_l, comp2_r = calc_bars_borders(image, sob_v, prow, px21, px22, region2,
                                                     refine=True,
                                                     plot=plot, lbarid=lbarid, rbarid=rbarid,
                                                     plot2=False)

                if np.any(np.isnan([comp2_l, comp2_r])):
                    self.logger.warning("converting NaN value, border of=%d", idx + 1)
                    comp2_l, comp2_r = comp_l, comp_r
                elif comp2_l > comp2_r:
                    # Not refining
                    self.logger.warning("computed left border of=%d greater than right border", idx + 1)
                    comp2_l, comp2_r = comp_l, comp_r

            # print('slit', lbarid, '-', rbarid, comp_l, comp_r)
            # print('pos1', comp_l, comp_r)
            # print('pos2', comp2_l, comp2_r)

            xpos1_virt, _ = dist.wcs_pvex(wcs, comp2_l + 1, ref_y_l_d)
            xpos2_virt, _ = dist.wcs_pvex(wcs, comp2_r + 1, ref_y_r_d)

            y1_virt = ref_y_l_v - slit_h_virt - slit_h_tol
            y2_virt = ref_y_r_v + slit_h_virt + slit_h_tol
            _, y1 = dist.wcs_exvp(wcs, xpos1_virt + 1, y1_virt)
            _, y2 = dist.wcs_exvp(wcs, xpos2_virt + 1, y2_virt)
            # print(comp2_l, comp2_r, y1 - 1, y2 - 1)
            cbb = BoundingBox.from_coordinates(comp2_l, comp2_r, y1 - 1, y2 - 1)
            slits_bb[lbarid] = cbb
            mask1[cbb.slice] = lbarid

        self.save_intermediate_array(mask1, 'mask_slit_computed.fits')
        return slits_bb


def compute_off_rotation(data, wcsa, csu_conf, slits_bb=None, rotaxis=(0, 0),
                         logger=None, debug_plot=False,
                         intermediate_results=True
                         ):
    # We can use a dictionary of slits_bb created by measuring the slits
    # in the image or we can use the slits given by the model of the CSU
    # which is correct around ~ 1.5 pix

    if logger is None:
        logger = logging.getLogger(__name__)

    swapped_code = (sys.byteorder == 'little') and '>' or '<'
    if data.dtype.byteorder == swapped_code:
        data = data.byteswap().newbyteorder()

    logger.info('we have %s slits', len(csu_conf.slits))
    refslits = [slit for slit in csu_conf.slits.values() if slit.target_type is TargetType.REFERENCE]
    logger.info('we have %s reference slits', len(refslits))

    slits = []
    p2 = []  # Pos REF VIRT
    p1 = []  # Pos REF DET
    q1 = []  # Pos Measured DET
    q2 = []  # Pos Measured VIRT

    for this in refslits:
        if slits_bb is not None:
            if this.idx not in slits_bb:
                logger.warning('slit %s not detected, skipping', this.idx)
                continue

            bb = slits_bb[this.idx]
        else:
            bb = this.bbox()
        region = bb.slice
        target_coordinates = this.target_coordinates
        res = comp_centroid(data, bb, debug_plot=debug_plot, plot_reference=target_coordinates)
        logger.debug('slit %s is formed by bars %s %s', this.idx, this.lbars_ids, this.rbars_ids)
        logger.debug('in slit %s, reference is %s', this.idx, target_coordinates)
        logger.debug('in slit %s, reference (v) is %s', this.idx, this.target_coordinates_v)

        if res is None:
            logger.warning('no object found in slit %s, skipping', this.idx)
            continue

        if debug_plot and intermediate_results:
            ax = res[2]
            ax.set_title('slit %s' % this.idx)
            plt.savefig('centroid_slit_%s.png' % this.idx)
            plt.close()

        m_x = res[0] + region[1].start
        m_y = res[1] + region[0].start

        logger.debug('in slit %s, object is (%s, %s)', this.idx, m_x, m_y)
        slits.append(this.idx)
        p1.append(this.target_coordinates)
        p2.append(this.target_coordinates_v)
        q1.append((m_x, m_y))

    logger.info('compute offset and rotation with %d points', len(p1))
    qc = QC.BAD

    if len(p2) == 0:
        logger.warning("can't compute offset and rotation with 0 points")
        offset = np.array([0.0, 0.0])
        rot = np.array([[1, 0], [0, 1]])
        angle = 0.0
    else:
        logger.debug('convert coordinates to virtual, ie, focal plane')
        q2 = dist.wcs_pix2virt(wcsa, q1, origin=0)
        logger.debug('converted virtual coordinates')
        for idx, c in zip(slits, q2):
            logger.debug('in slit %s, object (v) is %s', idx, c)
        # Move from objects to reference
        logger.debug('compute transform from measured objects to reference coordinates')
        offset, rot = fit_offset_and_rotation(q2, p2)
        logger.debug('rotation matrix')
        logger.debug('%s', rot)
        logger.debug('translation (with rotation around 0)')
        logger.debug('%s', offset)

        logger.debug('center of rotation (from CRPIX) is %s', rotaxis)
        logger.debug('translation (with rotation around rotaxis)')
        newoff = np.dot(rot, offset - rotaxis) + rotaxis
        logger.debug('%s', newoff)

        offset = newoff
        angle = math.atan2(rot[1, 0], rot[0, 0])
        angle = np.rad2deg(angle)
        qc = QC.GOOD
    logger.info('offset is %s', offset)
    logger.info('rot matrix is %s', rot)
    logger.info('rot angle %5.2f deg', angle)

    pq1 = np.subtract(p1, q1)
    pq2 = np.subtract(p2, q2)
    if len(pq1) != 0:
        logger.debug('MEAN of REF-MEASURED (ON DETECTOR) %s', pq1.mean(axis=0))
        logger.debug('MEAN pf REF-MEASURED (VIRT) %s', pq2.mean(axis=0))

    return offset, angle, qc, (slits, p1,p2, q1, q2)


def calc_bars_borders(image, sob, prow, px1, px2, regionw, h=16, refine=False,
                      plot=True, lbarid=0, rbarid=0, plot2=False):

    borders = [px1 - regionw, px1 + regionw, px2 - regionw, px2 + regionw,
               prow - h, prow + h]
    borders = np.clip(borders, 0, 2047)
    cc1l = borders[0]
    cc1r = borders[1]
    cc2l = borders[2]
    cc2r = borders[3]
    py1 = borders[4]
    py2 = borders[5]

    bb = BoundingBox.from_coordinates(cc1l, cc2r, py1, py2)

    lres = []

    # print(cc1l, cc2r, py1, py2)

    steps = range(-3, 3 + 1)
    scale = 3
    for s in steps:
        pr1 = prow + scale * s
        if pr1 < 0 or pr1 >= 2048:
            continue
        val = calc_borders_step(image, sob, pr1, cc1l, cc1r, refine=refine,
                                plot=plot, barid=lbarid, px=px1)
        if val is not None:
            lres.append(val)

    # Average left bar position
    if lres:
        row_l, col_l, colr_l = zip(*lres)
    else:
        row_l, col_l, colr_l = [], [], []
    if refine:
        comp_l = np.mean(colr_l)
    else:
        comp_l = np.median(col_l)

    rres = []
    scale = 3
    for s in steps:
        pr1 = prow + scale * s
        if pr1 < 0 or pr1 >= 2048:
            continue

        val = calc_borders_step(image, sob, pr1, cc2l, cc2r, sign=-1, refine=refine,
                                plot=plot, barid=rbarid, px=px2)
        if val is not None:
            rres.append(val)
    # Average left bar position
    if rres:
        row_r, col_r, colr_r = zip(*rres)
    else:
        row_r, col_r, colr_r = [], [], []

    if refine:
        comp_r = np.mean(colr_r)
    else:
        comp_r = np.median(col_r)

    if plot2:
        plt.imshow(image[bb.slice], extent=bb.extent)
        plt.scatter(colr_l, row_l, marker='+', color='black')
        plt.scatter(colr_r, row_r, marker='+', color='black')
        plt.axvline(px1, color='blue')
        plt.axvline(px2, color='blue')
        plt.axvline(comp_l, color='red')
        plt.axvline(comp_r, color='red')
        plt.show()

    return comp_l, comp_r


def calc_borders_step(image, sob, pr, ccl, ccr, sign=1, refine=False, plot=True, barid=0, px=0):
    cut1 = sob[pr, ccl:ccr + 1]
    mcut1 = image[pr, ccl:ccr + 1]

    res = feat.peak_local_max(sign * cut1, exclude_border=4, num_peaks=2)

    if len(res) != 0:
        res1 = res[:, 0]
    else:
        res1 = []

    if sign > 0:
        bar_s = 'l'
    else:
        bar_s = 'r'

    if plot:
        xdummy = np.arange(ccl, ccr + 1)
        plt.title("{}bar {}, prow={} px={}".format(bar_s, barid, pr, px))
        plt.plot(xdummy, sign * cut1)
        plt.axvline(px, color='g')
        ax1 = plt.gca()
        ax2 = ax1.twinx()
        ax2.plot(xdummy, mcut1, '--')
        for r in res1:
            plt.axvline(ccl + r, color='red')
        for r in res[:, 0]:
            plt.axvline(ccl + r, color='red', linestyle='--')
    xpeak = None
    if len(res1) != 0:
        idx = (sign * cut1)[res1].argmax()
        xpeak = ccl + res1[idx]

    if plot:
        if xpeak:
            plt.axvline(xpeak, color='black')
        plt.show()

    if xpeak:
        if refine:
            x_t, y_t = refine_peaks(sign * sob[pr], np.array([xpeak]), window_width=3)
            xpeak_ref = x_t
        else:
            xpeak_ref = xpeak
        return pr, xpeak, xpeak_ref
    else:
        return None



#
# Copyright 2008-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

from __future__ import division

import enum
import numpy

from numina.array.bbox import BoundingBox

EMIR_NBARS = 55

class CSUConf(object):
    """Information about the configuration of slits in the CSU"""
    def __init__(self, barmodel):
        self.name = 'CSU'
        self.conf_id = 'v1'
        self.conf_f = 'UNKNOWN'
        self.slits = {}
        # Indices
        self.LBARS = list(range(1, EMIR_NBARS + 1))
        self.RBARS = [(i + EMIR_NBARS) for i in self.LBARS]
        self.NBARS = EMIR_NBARS
        self.bars = barmodel
        # CSU is open if
        self.L_LIMIT = 6
        self.R_LIMIT = 2057
        # Generate default slits, without header information
        self._update_slits({})

    def is_open(self):
        lopen = all(self.bars[barid].x2 <= self.L_LIMIT for barid in self.LBARS)
        ropen = all(self.bars[barid].x1 >= self.R_LIMIT for barid in self.RBARS)

        return lopen and ropen

    def is_closed(self):
        return not self.is_open()

    def set_state_from_img(self, img):
        """Set state using a FITS image"""

        # FIXME: CSUCONFF is in primary
        hdr0 = img[0].header
        self.conf_f = hdr0['CSUCONFF']

        if 'MECS' in img:
            # Get header from extension
            hdr_csu = img['MECS'].header
        else:
            # Get header from main header
            hdr_csu = img[0].header
        self.set_state(hdr_csu)

    def set_state(self, hdr):
        """Read CSU information and slits from header"""

        # FIXME: this keyword in only in primary
        # it should be duplicated in MECS
        try:
            self.conf_f = hdr['CSUCONFF']
        except KeyError:
            pass
        # Read CSUPOS and set position in model
        # The bars
        for idx in self.bars:
            key = "CSUP{}".format(idx)
            # UNIT is mm

            # set CSUPOS for BAR
            # Using the model, this sets the X,Y coordinate
            self.bars[idx].csupos = hdr[key]

        self._update_slits(hdr)

    def _update_slits(self, hdr):
        """Recreate slits"""

        # Slits.
        # For the moment we will fuse only reference slits
        OUTPOS = -100
        # clean existing slits
        self.slits = {}
        mm = []
        for idx in self.LBARS:

            # References from header
            try:
                slit_t = hdr["SLIFL%d" % idx]
                target_type = TargetType(slit_t)
            except KeyError:
                target_type = TargetType.UNKNOWN

            xref = hdr.get("XRSLI%d" % idx, OUTPOS) - 1
            yref = hdr.get("YRSLI%d" % idx, OUTPOS) - 1
            target_coordinates = (xref, yref)

            xref = hdr.get("XVSLI%d" % idx, OUTPOS) - 1
            yref = hdr.get("YVSLI%d" % idx, OUTPOS) - 1
            target_coordinates_v = (xref, yref)

            mm.append((idx, target_type, target_coordinates, target_coordinates_v))

        bag = merge_slits(mm, max_slits=3, tol=1e-2)

        for idx, l_ids in bag.items():
            r_ids = [lid + EMIR_NBARS for lid in l_ids]

            lbars = {lid: self.bars[lid] for lid in l_ids}
            rbars = {rid: self.bars[rid] for rid in r_ids}

            this = LogicalSlit(idx, lbars=lbars, rbars=rbars)
            # References from header
            ref = mm[l_ids[0] - 1]
            this.target_type = ref[1]
            this.target_coordinates = ref[2]
            this.target_coordinates_v = ref[3]
            self.slits[idx] = this


class TargetType(enum.Enum):
    """Possible targets in a slit"""
    UNASSIGNED = 0
    SOURCE = 1
    REFERENCE = 2
    SKY = 3
    UNKNOWN = 4


class CSUBarModel(object):
    def __init__(self, idx, xs, xdelt, yc, active=True):
        self.idx = idx
        self.xs = xs
        self.xdelt = xdelt
        # FIXME: pixels?
        self.slit_h_virt = 16.242
        self.y0 = yc
        self.y1 = yc - self.slit_h_virt - 3
        self.y2 = yc + self.slit_h_virt + 3
        # Height in virt pixels
        self._x1 = 0
        self._x2 = 2048
        self.active = active
        self.csupos = 0.0

    def position(self, csupos):
        # start in 1 position
        return self.xs + self.xdelt * csupos

    @property
    def current_pos(self):
        # start in 1 position
        return self.position(self.csupos)

    @property
    def xpos(self):
        # start in 1 position
        return self.position(self.csupos)

    @property
    def x2(self):
        return self._x2

    @property
    def x1(self):
        return self._x1

    def bbox(self):
        # origin is 1
        x1 = max(self.x1 - 1, 0)
        y1 = max(self.y1 - 1, 0)
        x2 = max(self.x2 - 1, 0, x1)
        y2 = max(self.y2 - 1, 0, y1)
        return BoundingBox.from_coordinates(x1, x2, y1, y2)


class CSUBarModelL(CSUBarModel):

    @property
    def x2(self):
        return self.position(self.csupos)


class CSUBarModelR(CSUBarModel):

    @property
    def x1(self):
        return self.position(self.csupos)


class PhysicalBar(object):
    def __init__(self, idx, xpos, y1, y2, active=True):
        self.idx = idx
        self.xpos = xpos
        self.y1 = y1
        self.y2 = y2
        self.active = active


class PhysicalBarL(PhysicalBar):
    def __init__(self, idx, xpos, y1, y2, active=True):
        super(PhysicalBarL, self).__init__(idx, xpos, y1, y2, active)


class PhysicalBarR(PhysicalBar):
    def __init__(self, idx, xpos, y1, y2, active=True):
        super(PhysicalBarR, self).__init__(idx, xpos, y1, y2, active)


class LogicalSlit(object):
    """Slit formed from combination of PhysicalBarL and PhysicalBarR"""
    def __init__(self, idx, lbars, rbars, target_type=TargetType.UNKNOWN):
        UNDEFPOS = -100
        self.target_type = target_type
        self.target_coordinates = (UNDEFPOS, UNDEFPOS)
        self.target_coordinates_v = (UNDEFPOS, UNDEFPOS)
        self.idx = idx
        self.lbars = lbars
        self.rbars = rbars
        # Bar ids
        lbars_ids = list(lbars.keys())
        rbars_ids = list(rbars.keys())
        lbars_ids.sort()
        rbars_ids.sort()

        # Must have the same size
        if len(lbars) != len(rbars):
            raise ValueError('len() are different')

        # bars must be paired
        for l, r in zip(lbars_ids, rbars_ids):
            if r != l + EMIR_NBARS:
                raise ValueError('not paired {} and {}'.format(l, r))

        for a, b in zip(lbars_ids, lbars_ids[1:]):
            if b != a + 1:
                raise ValueError('not contiguous {} and {}'.format(a, b))

        self.lbars_ids = lbars_ids
        self.rbars_ids = rbars_ids
        # BBOX
        id_first = self.lbars_ids[0]
        id_last = self.lbars_ids[-1]

        bbox_y1 = self.lbars[id_first].y1
        bbox_y2 = self.lbars[id_last].y2

        bbox_x1 = min(self.lbars.values(), key=lambda obk: obk.xpos).xpos
        bbox_x2 = max(self.rbars.values(), key=lambda obk: obk.xpos).xpos
        # Origin is 1
        self.bbox_int = BoundingBox.from_coordinates(bbox_x1-1, bbox_x2-1, bbox_y1-1, bbox_y2-1)

    def bbox(self):
        return self.bbox_int

    def width(self):
        bar_x1 = min(self.lbars.values(), key=lambda obk: obk.xpos)
        bar_x2 = max(self.rbars.values(), key=lambda obk: obk.xpos)
        return bar_x2.xpos - bar_x1.xpos


def create_bar_models(barstab):
    """Create a dictionary of models of bars using a table"""
    bars = {}
    for params in barstab:
        barid = int(params[0])
        x0 = params[3]
        xdelt = params[2]
        y0 = params[1]
        if barid > EMIR_NBARS:
            bar = CSUBarModelR(barid, x0, xdelt, y0)
        else:
            bar = CSUBarModelL(barid, x0, xdelt, y0)
        bars[barid] = bar
    return bars


def read_csu_from_header(barmodel, hdr):
    """Read CSU information and slits from header"""
    # FIXME: CSUCONFF is in primary
    csu_conf = CSUConf(barmodel)
    csu_conf.set_state(hdr)
    return csu_conf


def read_csu_from_image(barmodel, img):
    """Read CSU information and slits from FITS"""
    csu_conf = CSUConf(barmodel)
    csu_conf.set_state_from_img(img)
    return csu_conf


def merge_slits(mm, max_slits=3, tol=1e-2):
    """"Merge contiguous REFERENCE slits"""
    cx1 = 0
    bag = {}
    while True:
        t_id, t_type, t_coor, t_coor_v = mm[cx1]
        rest = mm[cx1 + 1: cx1 + max_slits]
        bag[t_id] = [t_id]
        cx2 = cx1 + 1
        if t_type == TargetType.REFERENCE:
            # For REFERENCE
            for r_id, r_type, r_coor, r_coor_v in rest:
                dis = numpy.hypot(t_coor[0] - r_coor[0], t_coor[1] - r_coor[1])
                if r_type == TargetType.REFERENCE and (dis < tol) :
                    bag[t_id].append(r_id)
                    cx2 += 1
                else:
                    break
        cx1 = cx2
        if cx1 >= len(mm):
            break
    return bag



#
# Copyright 2016-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Spectroscopy mode, coadd ABBA
"""


import numina.core
import numina.array.combine as combine
from numina.core import Result
from numina.core.requirements import ObservationResultRequirement
import numina.exceptions
import numina.ext.gtc
import numina.core.query as qmod
from numina.processing.combine import basic_processing_with_combination

from emirdrp.core.recipe import EmirRecipe
import emirdrp.products as prods


class CoaddABBARecipe(EmirRecipe):
    """Process images in ABBA mode"""

    obresult = ObservationResultRequirement(
        query_opts=qmod.ResultOf(
            'LS_ABBA.reduced_mos_abba',
            node='children',
            id_field="stareSpectraIds"
        )
    )

    reduced_mos_abba = Result(prods.ProcessedMOS)

    def build_recipe_input(self, obsres, dal):
        if numina.ext.gtc.check_gtc():
            self.logger.debug('running in GTC environment')
            return self.build_recipe_input_gtc(obsres, dal)
        else:
            self.logger.debug('running outside of GTC environment')
            return super(CoaddABBARecipe, self).build_recipe_input(
                obsres, dal
            )

    def build_recipe_input_gtc(self, obsres, dal):
        self.logger.debug('start recipe input builder')
        stareImagesIds = obsres.stareSpectraIds
        self.logger.debug('ABBA images IDS %s: ', stareImagesIds)
        stareImages = []
        for subresId in stareImagesIds:
            subres = dal.getRecipeResult(subresId)
            stareImages.append(subres['elements']['reduced_mos_abba'])

        newOR = numina.core.ObservationResult()
        newOR.frames = stareImages
        newRI = self.create_input(obresult=newOR)
        self.logger.debug('end recipe input builder')
        return newRI

    def run(self, rinput):
        self.logger.info('starting spectroscopy ABBA coadd reduction')

        flow = self.init_filters(rinput)

        nimages = len(rinput.obresult.frames)
        self.logger.info('we receive %d images', nimages)
        if nimages == 0:
            msg = 'Received %d images' % nimages
            raise numina.exceptions.RecipeError(msg)

        hdulist = basic_processing_with_combination(
            rinput,
            flow,
            method=combine.mean,
            prolog="Process Coadd ABBA"
        )

        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        # Update SEC to 0
        # hdr['SEC'] = 0

        result = self.create_result(reduced_mos_abba=hdulist)
        self.logger.info('end spectroscopy ABBA coadd reduction')
        return result


class CoaddRecipe(EmirRecipe):
    """Generic Coadd Recipe"""

    obresult = ObservationResultRequirement()

    result_coadd = Result(prods.ProcessedMOS)

    def build_recipe_input(self, obsres, dal):
        if numina.ext.gtc.check_gtc():
            self.logger.debug('running in GTC environment')
            return self.build_recipe_input_gtc(obsres, dal)
        else:
            self.logger.debug('running outside of GTC environment')
            return super(CoaddRecipe, self).build_recipe_input(
                obsres, dal
            )

    def build_recipe_input_gtc(self, obsres, dal):
        self.logger.debug('start recipe input builder')

        # This depends on the RecipeResult
        result_field = 'reduced_mos_abba'

        stareImagesIds = obsres.stareSpectraIds
        self.logger.debug('Coadd images IDS %s: ', stareImagesIds)
        stareImages = []
        for subresId in stareImagesIds:
            subres = dal.getRecipeResult(subresId)
            stareImages.append(subres['elements'][result_field])

        newOR = numina.core.ObservationResult()
        newOR.frames = stareImages
        newRI = self.create_input(obresult=newOR)
        self.logger.debug('end recipe input builder')
        return newRI

    def run(self, rinput):
        self.logger.info('starting coadd reduction')

        flow = self.init_filters(rinput)

        nimages = len(rinput.obresult.frames)
        self.logger.info('we receive %d images', nimages)
        if nimages == 0:
            msg = 'Received %d images' % nimages
            raise numina.exceptions.RecipeError(msg)

        hdulist = basic_processing_with_combination(
            rinput,
            flow,
            method=combine.mean,
            prolog="Process Generic Coadd"
        )

        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        # Update SEC to 0
        # hdr['SEC'] = 0

        result = self.create_result(spec_coadd_abba=hdulist)
        self.logger.info('end coadd reduction')
        return result



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from copy import deepcopy
from datetime import datetime
import json
import os
import sys
from uuid import uuid4

from numina.tools.arg_file_is_new import arg_file_is_new

from emirdrp.core import EMIR_NBARS


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser()

    # positional arguments
    parser.add_argument("filename",
                        help="TXT with list of bounddict files",
                        type=argparse.FileType('rt'))
    parser.add_argument("--outfile", required=True,
                        help="Output merged JSON file",
                        type=lambda x: arg_file_is_new(parser, x))

    # optional arguments
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args()

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # initialize empty output
    bounddict = {}

    # read list of JSON files to be merged
    file_content = args.filename.read().splitlines()
    next_file_is_first = True
    for line in file_content:
        if len(line) > 0:
            if line[0] != '#':
                tmpfile = line.split()[0]
                if not os.path.isfile(tmpfile):
                    raise ValueError("File " + tmpfile + " not found!")
                tmpbounddict = json.loads(open(tmpfile).read())
                if next_file_is_first:
                    bounddict = deepcopy(tmpbounddict)
                    # update some values
                    bounddict['meta_info']['creation_date'] = \
                        datetime.now().isoformat()
                    bounddict['uuid'] = str(uuid4())
                    next_file_is_first = False
                else:
                    for islitlet in range(EMIR_NBARS):
                        cslitlet = "slitlet" + str(islitlet).zfill(2)
                        if cslitlet in tmpbounddict['contents']:
                            for dateobs in tmpbounddict['contents'][cslitlet]:
                                bounddict['contents'][cslitlet][dateobs] = \
                                    tmpbounddict['contents'][cslitlet][dateobs]

    # save merged JSON file
    with open(args.outfile.name, 'w') as fstream:
        json.dump(bounddict, fstream, indent=2, sort_keys=True)
        print('>>> Saving file ' + args.outfile.name)

if __name__ == "__main__":

    main()


#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
import json
import numpy as np
import os.path
import sys

from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.ximshow import ximshow
from numina.tools.arg_file_is_new import arg_file_is_new
from emirdrp.instrument.csu_configuration import CsuConfiguration
from emirdrp.tools.nscan_minmax_frontiers import nscan_minmax_frontiers

from emirdrp.tools.fit_boundaries import bound_params_from_dict
from emirdrp.tools.fit_boundaries import expected_distorted_frontiers
from emirdrp.tools.fit_boundaries import overplot_boundaries_from_params
from emirdrp.tools.fit_boundaries import overplot_frontiers_from_params
from emirdrp.tools.list_slitlets_from_string import list_slitlets_from_string

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES
from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2


def select_unrectified_slitlet(image2d, islitlet, csu_bar_slit_center,
                               params, parmodel, maskonly):
    """Returns image with the indicated slitlet (zero anywhere else).

    Parameters
    ----------
    image2d : numpy array
        Initial image from which the slitlet data will be extracted.
    islitlet : int
        Slitlet number.
    csu_bar_slit_center : float
        CSU bar slit center.
    params : :class:`~lmfit.parameter.Parameters`
        Parameters to be employed in the prediction of the distorted
        boundaries.
    parmodel : str
        Model to be assumed. Allowed values are 'longslit' and
        'multislit'.
    maskonly : bool
        If True, returns simply a mask (1 in the slitlet region and
        zero anywhere else.

    Returns
    -------
    image2d_output : numpy array
        2D image with the pixel information corresponding to the
        selected slitlet and zero everywhere else.

    """

    # protection
    if image2d.shape != (EMIR_NAXIS2, EMIR_NAXIS1):
        raise ValueError("NAXIS1, NAXIS2 unexpected for EMIR detector")

    # initialize image output
    image2d_output = np.zeros_like(image2d)

    # expected slitlet frontiers
    list_expected_frontiers = expected_distorted_frontiers(
        islitlet, csu_bar_slit_center,
        params, parmodel, numpts=101, deg=5, debugplot=0
    )
    pol_lower_expected = list_expected_frontiers[0].poly_funct
    pol_upper_expected = list_expected_frontiers[1].poly_funct

    # main loop: compute for each channel the minimum and maximum scan
    for j in range(EMIR_NAXIS1):
        xchannel = j + 1
        y0_lower = pol_lower_expected(xchannel)
        y0_upper = pol_upper_expected(xchannel)
        n1, n2 = nscan_minmax_frontiers(y0_frontier_lower=y0_lower,
                                        y0_frontier_upper=y0_upper,
                                        resize=True)
        # note that n1 and n2 are scans (ranging from 1 to NAXIS2)
        if maskonly:
            image2d_output[(n1 - 1):n2, j] = np.repeat(
                [1.0], (n2 - n1 + 1)
            )
        else:
            image2d_output[(n1 - 1):n2, j] = image2d[(n1 - 1):n2, j]

    return image2d_output


def main(args=None):

    # parse command-line options
    parser = argparse.ArgumentParser()

    # positional arguments
    parser.add_argument("fitsfile",
                        help="FITS file name to be displayed",
                        type=argparse.FileType('rb'))
    parser.add_argument("--fitted_bound_param", required=True,
                        help="JSON file with fitted boundary coefficients "
                             "corresponding to the multislit model",
                        type=argparse.FileType('rt'))
    parser.add_argument("--slitlets", required=True,
                        help="Slitlet selection: string between double "
                             "quotes providing tuples of the form "
                             "n1[,n2[,step]]",
                        type=str)

    # optional arguments
    parser.add_argument("--outfile",
                        help="Output FITS file name",
                        type=lambda x: arg_file_is_new(parser, x, mode='wb'))
    parser.add_argument("--maskonly",
                        help="Generate mask for the indicated slitlets",
                        action="store_true")
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting/debugging" +
                             " (default=0)",
                        type=int, default=0,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")

    args = parser.parse_args()

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # read input FITS file
    hdulist_image = fits.open(args.fitsfile.name)
    image_header = hdulist_image[0].header
    image2d = hdulist_image[0].data

    naxis1 = image_header['naxis1']
    naxis2 = image_header['naxis2']

    if image2d.shape != (naxis2, naxis1):
        raise ValueError("Unexpected error with NAXIS1, NAXIS2")

    if image2d.shape != (EMIR_NAXIS2, EMIR_NAXIS1):
        raise ValueError("NAXIS1, NAXIS2 unexpected for EMIR detector")

    # remove path from fitsfile
    if args.outfile is None:
        sfitsfile = os.path.basename(args.fitsfile.name)
    else:
        sfitsfile = os.path.basename(args.outfile.name)

    # check that the FITS file has been obtained with EMIR
    instrument = image_header['instrume']
    if instrument != 'EMIR':
        raise ValueError("INSTRUME keyword is not 'EMIR'!")

    # read GRISM, FILTER and ROTANG from FITS header
    grism = image_header['grism']
    spfilter = image_header['filter']
    rotang = image_header['rotang']

    # read fitted_bound_param JSON file
    fittedpar_dict = json.loads(open(args.fitted_bound_param.name).read())
    params = bound_params_from_dict(fittedpar_dict)
    if abs(args.debugplot) in [21, 22]:
        params.pretty_print()

    parmodel = fittedpar_dict['meta_info']['parmodel']
    if parmodel != 'multislit':
        raise ValueError("Unexpected parameter model: ", parmodel)

    # define slitlet range
    islitlet_min = fittedpar_dict['tags']['islitlet_min']
    islitlet_max = fittedpar_dict['tags']['islitlet_max']
    list_islitlet = list_slitlets_from_string(
        s=args.slitlets,
        islitlet_min=islitlet_min,
        islitlet_max=islitlet_max
    )

    # read CsuConfiguration object from FITS file
    csu_config = CsuConfiguration.define_from_fits(args.fitsfile)

    # define csu_bar_slit_center associated to each slitlet
    list_csu_bar_slit_center = []
    for islitlet in list_islitlet:
        list_csu_bar_slit_center.append(
            csu_config.csu_bar_slit_center(islitlet))

    # initialize output data array
    image2d_output = np.zeros((naxis2, naxis1))

    # main loop
    for islitlet, csu_bar_slit_center in \
            zip(list_islitlet, list_csu_bar_slit_center):
        image2d_tmp = select_unrectified_slitlet(
            image2d=image2d,
            islitlet=islitlet,
            csu_bar_slit_center=csu_bar_slit_center,
            params=params,
            parmodel=parmodel,
            maskonly=args.maskonly
        )
        image2d_output += image2d_tmp

    # update the array of the output file
    hdulist_image[0].data = image2d_output

    # save output FITS file
    hdulist_image.writeto(args.outfile)

    # close original image
    hdulist_image.close()

    # display full image
    if abs(args.debugplot) % 10 != 0:
        ax = ximshow(image2d=image2d_output,
                     title=sfitsfile + "\n" + args.slitlets,
                     image_bbox=(1, naxis1, 1, naxis2), show=False)

        # overplot boundaries
        overplot_boundaries_from_params(
            ax=ax,
            params=params,
            parmodel=parmodel,
            list_islitlet=list_islitlet,
            list_csu_bar_slit_center=list_csu_bar_slit_center
        )

        # overplot frontiers
        overplot_frontiers_from_params(
            ax=ax,
            params=params,
            parmodel=parmodel,
            list_islitlet=list_islitlet,
            list_csu_bar_slit_center=list_csu_bar_slit_center,
            micolors=('b', 'b'), linetype='-',
            labels=False    # already displayed with the boundaries
        )

        # show plot
        pause_debugplot(12, pltshow=True)


if __name__ == "__main__":

    main()



#
# Copyright 2016-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""DTU configuration class"""

from __future__ import division

import contextlib

import numpy
import astropy.wcs


@contextlib.contextmanager
def managed_ndig(obj, ndig):
    """Context manager to handle ndig"""
    old_ndig = obj.get_ndig()
    obj.set_ndig(ndig)
    try:
        yield ndig
    finally:
        # Code to release resource, e.g.:
        obj.set_ndig(old_ndig)


class DtuAxis(object):
    """Represents one DTU axis of movement"""
    def __init__(self, name, coor, coor_f=1.0, coor_0=0.0):
        if name.lower() not in ['x', 'y', 'z']:
            raise ValueError('"name" must be "X", "Y" or "Z')
        self.name = name
        self.coor = coor
        self.coor_f = coor_f
        self.coor_0 = coor_0
        self.ndig = 3

    @property
    def coor_r(self):
        return (self.coor / self.coor_f) - self.coor_0

    @classmethod
    def from_header(cls, hdr, name):
        if name.lower() not in ['x', 'y', 'z']:
            raise ValueError('"name" must be "X", "Y" or "Z')
        coor = hdr['{}DTU'.format(name)]
        coor_f = hdr.get('{}DTU_F'.format(name), 1.0)
        coor_0 = hdr.get('{}DTU_0'.format(name), 0.0)
        return DtuAxis(name, coor, coor_f, coor_0)

    def allclose(self, other, rtol=1e-05, atol=1e-08, equal_nan=False):
        import numpy
        a = [self.coor, self.coor_f, self.coor_0]
        b = [other.coor, other.coor_f, other.coor_0]
        return numpy.allclose(a, b, rtol=rtol, atol=atol, equal_nan=equal_nan)

    def closeto(self, other, abserror):
        return self.allclose(other, rtol=0.0, atol=abserror, equal_nan=False)

    def set_ndig(self, ndig):
        self.ndig = ndig

    def get_ndig(self):
        return self.ndig

    def __eq__(self, other):
        # note: set the precision (number of decimal places) to the same
        # number employed in __str__() function above to print out member
        # values
        # FIXME: this is eqv to allclose with atol=1e-4?
        if isinstance(other, DtuAxis):
            result = \
                self.name.lower() == other.name.lower() and \
                (round(self.coor, self.ndig) == round(other.coor, self.ndig)) and \
                (round(self.coor_f, self.ndig) == round(other.coor_f, self.ndig)) and \
                (round(self.coor_0, self.ndig) == round(other.coor_0, self.ndig))

            return result
        return NotImplemented

    def __ne__(self, other):
        return not self == other


class DtuConf(object):
    """Represents the three DTU axes of movement"""
    def __init__(self, xaxis, yaxis, zaxis):
        self.xaxis = xaxis
        self.yaxis = yaxis
        self.zaxis = zaxis

    def set_ndig(self, ndig):
        self.xaxis.set_ndig(ndig)
        self.yaxis.set_ndig(ndig)
        self.zaxis.set_ndig(ndig)

    def get_ndig(self):
        return self.xaxis.get_ndig()

    @property
    def xdtu_r(self):
        return self.xaxis.coor_r

    @property
    def ydtu_r(self):
        return self.yaxis.coor_r

    @property
    def zdtu_r(self):
        return self.zaxis.coor_r

    @property
    def coor_r(self):
        return [self.xaxis.coor_r, self.yaxis.coor_r, self.zaxis.coor_r]

    @property
    def coor(self):
        return [self.xaxis.coor, self.yaxis.coor, self.zaxis.coor]

    def allclose(self, other, rtol=1e-05, atol=1e-08, equal_nan=False):
        return (self.xaxis.allclose(other.xaxis, rtol=rtol, atol=atol, equal_nan=equal_nan) and
                self.yaxis.allclose(other.yaxis, rtol=rtol, atol=atol, equal_nan=equal_nan) and
                self.zaxis.allclose(other.zaxis, rtol=rtol, atol=atol, equal_nan=equal_nan)
                )

    def closeto(self, other, abserror):
        return self.allclose(other, rtol=0.0, atol=abserror, equal_nan=False)

    @classmethod
    def from_header(cls, hdr):
        """Create a DtuConf object from map-like header"""
        xaxis = DtuAxis.from_header(hdr, name='X')
        yaxis = DtuAxis.from_header(hdr, name='Y')
        zaxis = DtuAxis.from_header(hdr, name='Z')
        dtuconf = DtuConf(xaxis, yaxis, zaxis)
        return dtuconf

    @classmethod
    def from_img(cls, img):
        """Create a DtuConf object from a FITS image"""
        if 'MECS' in img:
            hdr = img['MECS'].header
        else:
            hdr = img[0].header

        return cls.from_header(hdr)

    def __eq__(self, other):
        if isinstance(other, DtuConf):
            return (self.xaxis == other.xaxis and
                    self.yaxis == other.yaxis and
                    self.zaxis == other.zaxis)

        return NotImplemented

    def __ne__(self, other):
        return not self == other

    def describe(self, ndig=None):
        # note: set the number of decimal figures in output to the precision
        # value employed in __eq__() function below
        if ndig is None:
            ndig = self.xaxis.ndig

        output = (
            "<DtuConf instance>\n"
              "- XDTU..: {0.xaxis.coor:{width}.{prec}f}\n"
              "- YDTU..: {0.yaxis.coor:{width}.{prec}f}\n"
              "- ZDTU..: {0.zaxis.coor:{width}.{prec}f}\n"
              "- XDTU_0: {0.xaxis.coor_0:{width}.{prec}f}\n"
              "- YDTU_0: {0.yaxis.coor_0:{width}.{prec}f}\n"
              "- ZDTU_0: {0.zaxis.coor_0:{width}.{prec}f}\n"
              "- XDTU_F: {0.xaxis.coor_f:{width}.{prec}f}\n"
              "- YDTU_F: {0.yaxis.coor_f:{width}.{prec}f}\n"
              "- ZDTU_F: {0.zaxis.coor_f:{width}.{prec}f}\n"
        )
        return output.format(self, width=8, prec=ndig)

    def to_wcs(self):
        """Create a WCS structure for DTU measurements"""
        dtur = self.coor_r
        xfac = dtur[0]
        yfac = -dtur[1]

        dtuwcs = astropy.wcs.WCS(naxis=2)
        dtuwcs.wcs.name = 'DTU WCS'
        dtuwcs.wcs.crpix = [0, 0]
        dtuwcs.wcs.cdelt = [1, 1]
        dtuwcs.wcs.crval = [yfac, xfac]
        dtuwcs.wcs.ctype = ['linear', 'linear']
        dtuwcs.wcs.cunit = ['um', 'um']

        return dtuwcs


def apply_on_axis(func, args):
    """Apply a function to the members of a sequence of DtuAxis"""
    coor_avg = float(func([arg.coor for arg in args]))
    coor_f_avg = float(func([arg.coor_f for arg in args]))
    coor_0_avg = float(func([arg.coor_0 for arg in args]))
    names = [arg.name for arg in args]
    if names:
        name = names[0]
    else:
        # empty list, fallback
        name = "X"
    return DtuAxis(name, coor=coor_avg, coor_f=coor_f_avg, coor_0=coor_0_avg)


def average(*args):
    """Return DtuConf instance with averaged values."""
    xaxis_avg = apply_on_axis(numpy.mean, [arg.xaxis for arg in args])
    yaxis_avg = apply_on_axis(numpy.mean, [arg.yaxis for arg in args])
    zaxis_avg = apply_on_axis(numpy.mean, [arg.zaxis for arg in args])

    return DtuConf(xaxis_avg, yaxis_avg, zaxis_avg)


def maxdiff(*args):
    """Return DtuConf instance with maximum differences."""
    xaxis_avg = apply_on_axis(numpy.ptp, [arg.xaxis for arg in args])
    yaxis_avg = apply_on_axis(numpy.ptp, [arg.yaxis for arg in args])
    zaxis_avg = apply_on_axis(numpy.ptp, [arg.zaxis for arg in args])

    return DtuConf(xaxis_avg, yaxis_avg, zaxis_avg)



#
# Copyright 2011-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


"""
Stare Image mode of EMIR
"""

import logging


from numina.array import combine
from numina.core import Result
from numina.core.query import Ignore
from numina.core.recipes import timeit
from numina.processing.combine import basic_processing_with_combination
from numina.util.context import manage_fits

from emirdrp.core.recipe import EmirRecipe
import emirdrp.core.extra as extra
import emirdrp.requirements as reqs
import emirdrp.products as prods
import emirdrp.processing.combine as comb
import emirdrp.decorators


_logger = logging.getLogger(__name__)


class StareImageRecipe2(EmirRecipe):
    """Process images in Stare Image Mode"""

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()

    result_image = Result(prods.ProcessedImage)

    def run(self, rinput):
        self.logger.info('starting stare image reduction (offline)')

        frames = rinput.obresult.frames

        with manage_fits(frames) as list_of:
            c_img = comb.combine_images(list_of, method=combine.mean, errors=False)

        flow = self.init_filters(rinput)

        # Correct Bias if needed
        # Correct Dark if needed
        # Correct FF

        processed_img = flow(c_img)

        hdr = processed_img[0].header
        self.set_base_headers(hdr)

        self.logger.debug('append BPM')
        if rinput.master_bpm is not None:
            self.logger.debug('using BPM from inputs')
            hdul_bpm = rinput.master_bpm.open()
            hdu_bpm = extra.generate_bpm_hdu(hdul_bpm[0])
        else:
            self.logger.debug('using empty BPM')
            hdu_bpm = extra.generate_empty_bpm_hdu(processed_img[0])

        # Append the BPM to the result
        processed_img.append(hdu_bpm)
        self.logger.info('end stare image (off) reduction')
        result = self.create_result(result_image=processed_img)

        return result


    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr = super(StareImageRecipe2, self).set_base_headers(hdr)
        # Set EXP to 0
        hdr['EXP'] = 0
        return hdr


class StareImageBaseRecipe(EmirRecipe):
    """Process images in Stare Image Mode"""

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    master_sky = reqs.MasterSkyRequirement(optional=True)

    frame = Result(prods.ProcessedImage)

    def __init__(self, *args, **kwargs):
        super(StareImageBaseRecipe, self).__init__(*args, **kwargs)
        if False:
            self.query_options['master_sky'] = Ignore()

    @emirdrp.decorators.loginfo
    @timeit
    def run(self, rinput):
        self.logger.info('starting stare image reduction')

        flow = self.init_filters(rinput)

        hdulist = basic_processing_with_combination(
            rinput,
            flow,
            method=combine.median
        )
        hdr = hdulist[0].header
        self.set_base_headers(hdr)

        if rinput.master_bpm:
            hdul_bpm = rinput.master_bpm.open()
            hdu_bpm = extra.generate_bpm_hdu(hdul_bpm[0])
        else:
            hdu_bpm = extra.generate_empty_bpm_hdu(hdulist[0])

        # Append the BPM to the result
        hdulist.append(hdu_bpm)
        self.logger.info('end stare image reduction')
        result = self.create_result(frame=hdulist)

        return result

    def set_base_headers(self, hdr):
        """Set metadata in FITS headers."""
        hdr = super(StareImageBaseRecipe, self).set_base_headers(hdr)
        # Update EXP to 0
        hdr['EXP'] = 0
        return hdr



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


"""Recipes for finding the best focus."""

from numina.core import Parameter
from numina.core import Result

import emirdrp.requirements as reqs
import emirdrp.products as prods
from emirdrp.core.recipe import EmirRecipe


class TelescopeRoughFocusRecipe(EmirRecipe):
    """Recipe to compute the telescope focus.

    **Observing modes:**

     * Telescope rough focus
     * Emir focus control

    **Inputs:**

     * A list of images
     * A list of sky images
     * Bias, dark, flat
     * A model of the detector
     * List of focii

    **Outputs:**
     * Best focus
    """

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    objects = Parameter([], 'List of x-y pair of object coordinates'),
    focus_range = Parameter([], 'Focus range: begin, end and step')

    focus = Result(prods.TelescopeFocus)

    def run(self, rinput):
        return self.create_result(focus=prods.TelescopeFocus())


class TelescopeFineFocusRecipe(EmirRecipe):
    """
    Recipe to compute the telescope focus.

    **Observing modes:**

        * Telescope fine focus

    **Inputs:**

     * A list of images
     * A list of sky images
     * Bias, dark, flat
     * A model of the detector
     * List of focii

    **Outputs:**
     * Best focus

    """

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    objects = Parameter([], 'List of x-y pair of object coordinates'),

    focus = Result(prods.TelescopeFocus)

    def run(self, rinput):
        return self.create_result(focus=prods.TelescopeFocus())


class DTUFocusRecipe(EmirRecipe):
    """
    Recipe to compute the DTU focus.

    **Observing modes:**

        * EMIR focus control

    **Inputs:**

     * A list of images
     * A list of sky images
     * Bias, dark, flat
     * A model of the detector
     * List of focii

    **Outputs:**
     * Best focus

    """

    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()
    objects = Parameter([], 'List of x-y pair of object coordinates'),
    msm_pattern = Parameter([], 'List of x-y pair of slit coordinates'),
    dtu_focus_range = Parameter([], 'Focus range of the DTU: begin, end and step')

    focus = Result(prods.DTUFocus)


    def run(self, rinput):
        return self.create_result(focus=prods.DTUFocus())



#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Auxiliary functions to rescale numpy arrays"""

from __future__ import division, print_function

import numpy as np


def rescale_array_to_z1z2(array, z1z2=(-1.0, 1.0)):
    """Rescale the values in a numpy array to the [z1,z2] interval.

    The transformation is carried out following the relation
    array_rs = b_flux * array - c_flux
    as explained in Appendix B1 of Cardiel (2009, MNRAS, 396, 680)

    Parameters
    ----------
    array : numpy array
        Numpy array to be rescaled.
    z1z2 : tuple, floats
        Minimum and maximum values in the returned array.

    Returns
    -------
    array_rs : numpy array
        Array with rescaled values.
    coef_rs : tuple, floats
        Coefficients b_flux and c_flux employed in the rescaling
        operation.

    """

    if type(array) is not np.ndarray:
        raise ValueError("array=" + str(array) + " must be a numpy.ndarray")

    array_min = array.min()
    array_max = array.max()

    z1, z2 = z1z2
    delta = array_max - array_min
    b_flux = (z2 - z1) / delta
    c_flux = (z2 * array_min - z1 * array_max) / delta

    array_rs = b_flux * array - c_flux

    return array_rs, (b_flux, c_flux)


def rescale_array_from_z1z2(array_rs, coef_rs=None):
    """Restore the values in a numpy array rescaled to the [z1,z2] interval.

    The transformation is carried out following the relation
    array = (array_rs + c_flux)/b_flux
    as explained in Appendix B1 of Cardiel (2009, MNRAS, 396, 680)

    Parameters
    ----------
    array_rs : numpy array
        Numpy array previously rescaled to the [z1,z2] interval with
        the function rescale_array_to_z1z2().
    coef_rs : tuple, floats
        Coefficients b_flux and c_flux previously employed in the
        rescaling operation. This tuple is one of the parameters
        returned by function_rescale_array_to_z1z2().

    Returns
    -------
    array : numpy array
        Array with restored values.

    """

    if type(array_rs) is not np.ndarray:
        raise ValueError(
            "array_rs=" + str(array_rs) + "must be a numpy.ndarray")

    b_flux, c_flux = coef_rs

    array = (array_rs + c_flux) / b_flux

    return array



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import re


def list_slitlets_from_string(s, islitlet_min, islitlet_max):
    """Return list of slitlets from string specification.

    Parameters
    ----------
    s : string
        String defining the slitlets. The slitlets must be specify
        as a set of n1[,n2[,step]] tuples. If only n1 is provided,
        the slitlet number n1 is considered. When n1 and n2 are given
        but step is missing, step=1 is assumed. Finally, when all
        n1, n2 and step are given, slitlets considered are those
        returned by range(n1, n2 + 1, step) function.
    islitlet_min : int
        Minimum slitlet number allowed.
    islitlet_max : int
        Maximum slitlet number allowed.

    Returns
    -------
    list_slitlets : Python list
        List of slitlets.

    """

    # protection
    if not isinstance(s, str):
        print('type(s): ', type(s))
        print('ERROR: function expected a string parameter')

    # initialize empty output
    set_slitlets = set()

    # remove leading blank spaces
    s = re.sub('^ *', '', s)
    # remove trailing blank spaces
    s = re.sub(' *$', '', s)
    # remove blank space before ','
    s = re.sub(' *,', ',', s)
    # remove blank spaces after  ','
    s = re.sub(', *', ',', s)
    # remove many blank spaces by a single blank space
    s = re.sub(' +', ' ', s)
    stuples = s.split()
    for item in stuples:
        subitems = item.split(',')
        nsubitems = len(subitems)
        if nsubitems == 1:
            n1 = int(subitems[0])
            n2 = n1
            step = 1
        elif nsubitems == 2:
            n1 = int(subitems[0])
            n2 = int(subitems[1])
            step = 1
        elif nsubitems == 3:
            n1 = int(subitems[0])
            n2 = int(subitems[1])
            step = int(subitems[2])
        else:
            raise ValueError('Unexpected slitlet range:', s)
        for i in range(n1, n2 + 1, step):
            if islitlet_min <= i <= islitlet_max:
                set_slitlets.add(i)
            else:
                print('islitlet_min: ', islitlet_min)
                print('islitlet_max: ', islitlet_max)
                print('i...........: ', i)
                raise ValueError("Slitlet number out of range!")

    list_slitlets = list(set_slitlets)
    list_slitlets.sort()

    # return result
    return list_slitlets



#
# Copyright 2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Generate YAML files to reduce ABBA observations"""

import argparse
import sys
import textwrap

from numina.array.display.fileinfo import FileInfo
from numina.tools.arg_file_is_new import arg_file_is_new


# auxiliary function to generate content of YAML files
def generate_yaml_content(args, list_fileinfo, enabled=True):

    step = args.step

    nimages = len(list_fileinfo)

    output = ''
    if step in [0, 1]:
        # initial reduction with STARE_IMAGE
        for i in range(nimages):
            if i % args.repeat == 0:
                if i != 0:
                    output += '---\n'
                idlabel = list_fileinfo[i].filename[:10]
                output += 'id: _{}'.format(idlabel)
                output += '\n'
                output += 'instrument: EMIR\n'
                output += 'mode: STARE_IMAGE\n'
                output += 'frames:\n'
            output += ' - ' + list_fileinfo[i].filename + '\n'
            if (i + 1) % args.repeat == 0:
                if step == 0:
                    output += 'enabled: True\n'
                elif step == 1:
                    output += 'enabled: False\n'
                else:
                    raise ValueError('Unexpected step={}'.format(step))

    if step == 1:
        # combination with FULL_DITHERED_IMAGE
        output += '---\n'
        if args.obsid_combined is None:
            output += 'id: _combined' + '\n'
        else:
            output += 'id: _' + args.obsid_combined + '\n'
        output += 'instrument: EMIR\n'
        output += 'mode: FULL_DITHERED_IMAGE\n'
        output += 'children:\n'
        for i in range(nimages):
            if i % args.repeat == 0:
                idlabel = list_fileinfo[i].filename[:10]
                output += ' - _' + idlabel + '\n'
        output += 'requirements:\n'
        output += '  iterations: 0\n'
        output += '  sky_images: 0\n'
        output += '  refine_offsets: False\n'
        output += 'enabled: True\n'

    return output


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: generate observation result YAML file',
    )

    # positional or required arguments
    parser.add_argument("filename",
                        help="TXT file with list of ABBA FITS files",
                        type=argparse.FileType('rt'))
    parser.add_argument("--step", required=True,
                        help=textwrap.dedent("""\
                        0: preliminary STARE_IMAGE
                        1: combination with FULL_DITHERED_IMAGE"""),
                        type=int, choices=[0, 1])
    parser.add_argument("--outfile", required=True,
                        help="Output YAML file name",
                        type=lambda x: arg_file_is_new(parser, x))

    # optional arguments
    parser.add_argument("--repeat",
                        help="Repetitions at each position",
                        default=1, type=int)
    parser.add_argument("--obsid_combined",
                        type=str)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31mExecuting: ' + ' '.join(sys.argv) + '\033[0m\n')

    # read TXT file
    with args.filename as f:
        file_content = f.read().splitlines()
    list_fileinfo = []
    for line in file_content:
        if len(line) > 0:
            if line[0] not in ['#', '@']:
                tmplist = line.split()
                tmpfile = tmplist[0]
                if len(tmplist) > 1:
                    tmpinfo = tmplist[1:]
                else:
                    tmpinfo = None
                list_fileinfo.append(FileInfo(tmpfile, tmpinfo))

    # check consistency of pattern, repeat and number of images
    nimages = len(list_fileinfo)
    if nimages % args.repeat != 0:
        raise ValueError('Unexpected number of images')

    output = generate_yaml_content(args, list_fileinfo)

    # generate YAML file
    with args.outfile as f:
        f.write(output)
    print('--> File {} generated!'.format(args.outfile.name))


if __name__ == "__main__":
    main()
